# distutils: include_dirs = dimod/include/

# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

__all__ = ['Vartype', 'vartype_info']


cdef extern from "dimod/vartypes.h" namespace "dimod" nogil:
    enum Vartype:
        BINARY
        SPIN
        INTEGER
        REAL

    cdef cppclass vartype_info [Bias]:
        @staticmethod
        Bias default_max(Vartype)
        @staticmethod
        Bias default_min(Vartype)
        @staticmethod
        Bias max(Vartype)
        @staticmethod
        Bias min(Vartype)



# distutils: language = c++
# cython: language_level=3

# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# cimport numpy as np

from dimod.libcpp.constrained_quadratic_model cimport ConstrainedQuadraticModel as cppConstrainedQuadraticModel
from dimod.constrained.cyexpression cimport cyObjectiveView, cyConstraintView
from dimod.cyqmbase.cyqmbase_float64 cimport cyQMBase_float64, bias_type, index_type
from dimod.cyvariables cimport cyVariables


cdef class cyConstrainedQuadraticModel:    
    cdef cppConstrainedQuadraticModel[bias_type, index_type] cppcqm
    
    cdef readonly cyObjectiveView objective
    """Objective to be minimized."""

    cdef readonly cyVariables variables
    """Variables in use over the objective and all constraints."""

    cdef readonly cyVariables constraint_labels
    """The labels for each of the constraints."""

    cdef readonly object dtype
    """The type of the biases."""

    cdef readonly object index_dtype
    """The type of the indices."""

    cdef public bint REAL_INTERACTIONS

    cdef readonly object constraints
    """Constraints as a mapping.

    This dictionary and its contents should not be modified.
    """

cdef object make_cqm(cppConstrainedQuadraticModel[bias_type, index_type] cppcqm)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import warnings

from typing import Callable, Collection, List, Mapping, Optional

import numpy.random

from numpy import prod
from dimod.binary_quadratic_model import BinaryQuadraticModel
from dimod.decorators import graph_argument
from dimod.typing import GraphLike, Variable
from dimod.vartypes import SPIN

__all__ = ['frustrated_loop']


Predicate = Callable[[List[Variable]], bool]


@graph_argument('graph')
def frustrated_loop(graph: GraphLike,
                    num_cycles: int,
                    R: float = float('inf'),
                    cycle_predicates: Collection[Predicate] = tuple(),
                    max_failed_cycles: int = 100,
                    plant_solution: bool = True,
                    planted_solution: Optional[Mapping[Variable, int]] = None,
                    seed: Optional[int] = None,
                    ) -> BinaryQuadraticModel:
    """Generate a frustrated-loop problem.

    A generic frustrated-loop (FL) problem is a sum of Hamiltonians, each 
    generated from a single loop, as follows:

    1. Generate a loop by random walking on the specified graph, ``graph``.
    2. If the cycle meets provided predicates, continue; if not, return to  step 1.
    3. Choose one edge of the loop uniformly at random to be anti-ferromagnetic 
       (AFM, coupling value of `1`) and all others ferromagnetic 
       (`-1`). Or if :code:`plant_solution is False`, sample uniformly couplers
       subject to the constraint that there are an odd number of AFM couplers.
    4. Add the loop's interactions to the FL problem.
       If at any time the absolute value of an interaction in the FL problem,
       accumulated on an edge over good loops, exceeds a given maximum, ``R``,
       remove that edge from consideration in the loop generation procedure.

    This is a generic generator of FL problems that encompasses both the original
    FL problem definition [#HJARTL]_ and the limited FL problem
    definition [#KLH]_.

    Args:
        graph:
            The graph to build the frustrated loops on. Either an integer, `n`,
            interpreted as a complete graph of size `n`, a nodes/edges pair, a
            list of edges, or a NetworkX graph.

        num_cyles:
            Desired number of frustrated cycles.

        R:
            Maximum absolute interaction weight an edge can accumulate from good
            cycles.

        cycle_predicates:
            An iterable of functions, which should accept a cycle and return a bool.

        max_failed_cycles:
            Maximum number of failures to find a cycle before terminating.

        plant_solution: 
            Select frustrated loops with only 1 AFM coupler all 1 (and all -1) 
            solutions are among the ground states. 

        planted_solution:
            A dictionary assigning variables to spin states.  When provided, and
            :code:`planted_solution=True` spin states are relabeled so that the
            variable assignment becomes a planted ground state in place of the 
            all 1 solution.  
            This option is deprecated; use of spin-reversal transforms is 
            recommended for this purpose.

        seed: Random seed.

    .. [#HJARTL] Hen, I., J. Job, T. Albash, T.F. Rønnow, M. Troyer, D. Lidar. Probing for quantum
        speedup in spin glass problems with planted solutions. https://arxiv.org/abs/1502.01663v2

    .. [#KLH] King, A.D., T. Lanting, R. Harris. Performance of a quantum annealer on range-limited
        constraint satisfaction problems. https://arxiv.org/abs/1502.02098

    """
    nodes, edges = graph
    if num_cycles <= 0:
        raise ValueError("num_cycles should be a positive integer")
    if R <= 0:
        raise ValueError("R should be a positive integer")
    if max_failed_cycles <= 0:
        raise ValueError("max_failed_cycles should be a positive integer")
    
    r = numpy.random.RandomState(seed)

    adj = {v: set() for v in nodes}
    for u, v in edges:
        if u in adj:
            adj[u].add(v)
        else:
            adj[u] = {v}
        if v in adj:
            adj[v].add(u)
        else:
            adj[v] = {u}
    bqm = BinaryQuadraticModel({v: 0.0 for v in nodes}, {edge: 0.0 for edge in edges}, 0.0, SPIN)

    failed_cycles = 0
    good_cycles = 0
    while good_cycles < num_cycles and failed_cycles < max_failed_cycles:

        cycle = _random_cycle(adj, r)

        # if the cycle failed or it is otherwise invalid, mark as failed and continue
        if cycle is None or not all(pred(cycle) for pred in cycle_predicates):
            failed_cycles += 1
            continue
        good_cycles += 1

        if plant_solution:
            # randomly select from all frustrated loops with one AFM edge
            cycle_J = {(cycle[i - 1], cycle[i]) : -1 for i in range(len(cycle))}
            idx = r.randint(len(cycle))
            cycle_J[(cycle[idx - 1], cycle[idx])] = 1
        else:
            # randomly select from all frustrated loops (odd number of AFM edges)
            cycle_J = {(cycle[i], cycle[i+1]) : -1 for i in range(len(cycle)-1)}
            cycle_J[(cycle[-1],cycle[0])] = (1 - 2*(len(cycle_J) & 1))*prod(list(cycle_J.values()))
            
        # update the bqm
        bqm.add_interactions_from(cycle_J)
        for u, v in cycle_J:
            if abs(bqm.adj[u][v]) >= R:
                adj[u].remove(v)
                adj[v].remove(u)

    if good_cycles < num_cycles:
        raise RuntimeError(f"number of found cycles, {good_cycles}, is below " 
                           f"requested {num_cycles}")

    if planted_solution is not None:
        warnings.warn("planted_solution is deprecated; use spin-reversal transforms: "
                      "(1) bqm.flip_variable() for superficial reorientation of "
                      "spin labels to match desired planting or "
                      "(2) SpinReversalComposite for presentation of a randomized "
                      "planted solution to some solver.",
                      DeprecationWarning, stacklevel=3)
        # A spin-reversal transform for a BQM with zero linear biases
        for e in bqm.quadratic:
            bqm.set_quadratic(e[0], e[1],
                              bqm.quadratic[e]*planted_solution[e[0]]*planted_solution[e[1]])

    return bqm


def _random_cycle(adj, random_state):
    """Find a cycle using a random graph walk."""

    # step through idx values in adj to pick a random one, random.choice does not work on dicts
    n = random_state.randint(len(adj))
    for idx, v in enumerate(adj):
        if idx == n:
            break
    start = v

    walk = [start]
    visited = {start: 0}

    while True:
        if len(walk) > 1:
            # as long as we don't step back one we won't have any repeated edges
            previous = walk[-2]
            neighbors = [u for u in adj[walk[-1]] if u != previous]
        else:
            neighbors = list(adj[walk[-1]])

        if not neighbors:
            # we've walked into a dead end
            return None

        # get a random neighbor
        u = random_state.choice(neighbors)
        if u in visited:
            # if we've seen this neighbour, then we have a cycle starting from it
            return walk[visited[u]:]
        else:
            # add to walk and keep moving
            walk.append(u)
            visited[u] = len(visited)



# distutils: include_dirs = dimod/include/

# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from libcpp.string cimport string
from libcpp.vector cimport vector

from dimod.libcpp.constrained_quadratic_model cimport ConstrainedQuadraticModel

__all__ = ['LPModel', 'read']


cdef extern from "dimod/lp.h" namespace "dimod::lp" nogil:
    cdef cppclass LPModel[Bias, Index]:
        ConstrainedQuadraticModel[Bias, Index] model
        vector[string] variable_labels
        vector[string] constraint_labels

    LPModel[Bias, Index] read[Bias, Index] (const string) except +



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import Hashable, FrozenSet, List, Mapping, Optional, Tuple, Union
from numbers import Number

import itertools
import warnings

from collections import Counter

from collections import defaultdict
from functools import partial

import numpy as np

import dimod
from dimod.binary_quadratic_model import BinaryQuadraticModel
from dimod.constrained import ConstrainedQuadraticModel
from dimod.higherorder.polynomial import BinaryPolynomial
from dimod.sampleset import as_samples
from dimod.typing import Bias, Polynomial, SamplesLike, Variable
from dimod.vartypes import as_vartype, Vartype

__all__ = ['make_quadratic', 'make_quadratic_cqm', 'reduce_binary_polynomial']


def _spin_product(variables):
    """A BQM with a gap of 1 that represents the product of two spin variables.

    Note that spin-product requires an auxiliary variable.

    Args:
        variables (list):
            multiplier, multiplicand, product, aux

    Returns:
        :obj:`.BinaryQuadraticModel`

    """
    multiplier, multiplicand, product, aux = variables

    return BinaryQuadraticModel({multiplier: -.5,
                                 multiplicand: -.5,
                                 product: -.5,
                                 aux: -1.},
                                {(multiplier, multiplicand): .5,
                                 (multiplier, product): .5,
                                 (multiplier, aux): 1.,
                                 (multiplicand, product): .5,
                                 (multiplicand, aux): 1.,
                                 (product, aux): 1.},
                                2.,
                                Vartype.SPIN)

def _new_product(variables, u, v):
    # make a new product variable not in variables, then add it
    p = '{}*{}'.format(u, v)
    while p in variables:
        p = '_' + p
    variables.add(p)
    return p


def _new_aux(variables, u, v):
    # make a new auxiliary variable not in variables, then add it
    aux = 'aux{},{}'.format(u, v)
    while aux in variables:
        aux = '_' + aux
    variables.add(aux)
    return aux


def _decrement_count(idx, que, pair):
    count = len(idx[pair])
    que_count = que[count]
    que_count.remove(pair)
    if not que_count:
        del que[count]
    if count > 1:
        que[count - 1].add(pair)


def _remove_old(idx, term, pair):
    idx_pair = idx[pair]
    del idx_pair[term]
    if not idx_pair:
        del idx[pair]


def reduce_binary_polynomial(poly: BinaryPolynomial) -> Tuple[
        List[Tuple[FrozenSet[Hashable], Number]],
        List[Tuple[FrozenSet[Hashable], Hashable]]
    ]:
    """Reduce a binary polynomial to linear and quadratic terms, plus constraints.

    Introduces auxillary variables and constraints to reduce the polynomial
    to linear and quadratic terms.

    Args:
        poly: a binary polynomial that might have higher order terms.

    Returns:
        Two-tuple of a list of terms and their biases, as tuples, and a list of
        the original and auxiliary variables, as a tuple.

    Example:
        >>> poly = dimod.BinaryPolynomial({(0,): -1, (1,): 1, (2,): 1.5, (0, 1): -1, (0, 1, 2): -2}, dimod.BINARY)
        >>> dimod.reduce_binary_polynomial(poly)           # doctest:+SKIP
        ([(frozenset({0}), -1),
          (frozenset({1}), 1),
          (frozenset({2}), 1.5),
          (frozenset({0, 1}), -1),
          (frozenset({'0*1', 2}), -2)],
         [(frozenset({0, 1}), '0*1')])
    """

    variables = poly.variables
    constraints = []

    reduced_terms = []
    idx = defaultdict(dict)
    for item in poly.items():
        term, bias = item
        if len(term) <= 2:
            reduced_terms.append(item)
        else:
            for pair in itertools.combinations(term, 2):
                idx[frozenset(pair)][term] = bias

    que = defaultdict(set)
    for pair, terms in idx.items():
        que[len(terms)].add(pair)

    while idx:
        new_pairs = set()
        most = max(que)
        que_most = que[most]
        pair = que_most.pop()
        if not que_most:
            del que[most]
        terms = idx.pop(pair)

        prod_var = _new_product(variables, *pair)
        constraints.append((pair, prod_var))
        prod_var_set = {prod_var}

        for old_term, bias in terms.items():
            common_subterm = (old_term - pair)
            new_term = common_subterm | prod_var_set

            for old_pair in map(frozenset, itertools.product(pair, common_subterm)):
                _decrement_count(idx, que, old_pair)
                _remove_old(idx, old_term, old_pair)

            for common_pair in map(frozenset, itertools.combinations(common_subterm, 2)):
                idx[common_pair][new_term] = bias
                _remove_old(idx, old_term, common_pair)

            if len(new_term) > 2:
                for new_pair in (frozenset((prod_var, v)) for v in common_subterm):
                    idx[new_pair][new_term] = bias
                    new_pairs.add(new_pair)
            else:
                reduced_terms.append((new_term, bias))

        for new_pair in new_pairs:
            que[len(idx[new_pair])].add(new_pair)

    return reduced_terms, constraints


def _init_quadratic_model(qm, vartype, qm_factory):
    if vartype is None:
        if qm is None:
            raise ValueError("one of vartype or qm must be provided")
        else:
            vartype = qm.vartype
    else:
        vartype = as_vartype(vartype)  # handle other vartype inputs
        if qm is None:
            qm = qm_factory(vartype)
        else:
            qm = qm.change_vartype(vartype, inplace=False)

    # for backwards compatibility, add an info field
    if not hasattr(qm, 'info'):
        qm.info = {}
    qm.info['reduction'] = {}
    return qm, vartype

def _init_binary_polynomial(poly, vartype):
    if not (isinstance(poly, BinaryPolynomial) and (vartype in (poly.vartype, None))):
        poly = BinaryPolynomial(poly, vartype=vartype)
    return poly

def _init_objective(bqm, reduced_terms):
    for term, bias in reduced_terms:
        if len(term) == 2:
            bqm.add_interaction(*term , bias)
        elif len(term) == 1:
            bqm.add_variable(*term, bias)
        elif len(term) == 0:
            bqm.offset += bias
        else:
            # still has higher order terms, this shouldn't happen
            msg = ('Internal error: not all higher-order terms were reduced. '
                   'Please file a bug report.')
            raise RuntimeError(msg)

def make_quadratic_cqm(poly: Union[Polynomial, BinaryPolynomial],
                       vartype: Optional[Vartype] = None,
                       cqm: Optional[ConstrainedQuadraticModel] = None
                       ) -> ConstrainedQuadraticModel:
    """Create a constrained quadratic model from a higher order polynomial.

    Args:
        poly:
            Either a polynomial, as a dict of form `{term: bias, ...}`, where `term`
            is a tuple of one or more variables and `bias` the associated bias,
            or a :class:`.BinaryPolynomial`.

        vartype:
            Variable type for the binary quadratic model. Accepted input values:

            * :class:`~dimod.Vartype.SPIN`, ``'SPIN'``, ``{-1, 1}``
            * :class:`~dimod.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``

            If ``poly`` is a :class:`.BinaryPolynomial` , ``vartype`` is not required.

        cqm:
            Terms of the reduced polynomial are added to this constrained quadratic
            model. If not provided, a new constrained quadratic model is created.

    Examples:

        >>> poly = {(0,): -1, (1,): 1, (2,): 1.5, (0, 1): -1, (0, 1, 2): -2}
        >>> cqm = dimod.make_quadratic_cqm(poly, dimod.SPIN)

    """
    if not (vartype or isinstance(poly, BinaryPolynomial)):
        raise ValueError("can not infer vartype")
    cqm = cqm or ConstrainedQuadraticModel()
    vartype = vartype or poly.vartype
    poly = _init_binary_polynomial(poly, vartype)
    reduced_terms, constraints = reduce_binary_polynomial(poly)

    def var(x):
        return BinaryQuadraticModel({x: 1.0}, {}, 0.0, vartype)

    for (u, v), p in constraints:
        cqm.add_constraint( var(u)*var(v) - var(p)  == 0, label = f"'{u}'*'{v}' == '{p}'")

    obj = BinaryQuadraticModel(vartype=vartype)
    _init_objective(obj, reduced_terms)
    cqm.set_objective(obj + cqm.objective)

    return cqm


def make_quadratic(poly: Union[Polynomial, BinaryPolynomial], strength: float,
                   vartype: Optional[Vartype] = None,
                   bqm: Optional[BinaryQuadraticModel] = None) -> BinaryQuadraticModel:
    """Create a binary quadratic model from a higher order polynomial.

    Args:
        poly:
            Either a polynomial, as a dict of form `{term: bias, ...}`, where `term`
            is a tuple of one or more variables and `bias` the associated bias,
            or a :class:`.BinaryPolynomial`.

        strength:
            Energy penalty for violating the product constraint.
            Insufficient strength can result in the binary quadratic model not
            having the same minimizations as the polynomial.

        vartype:
            Variable type for the binary quadratic model. Accepted input values:

            * :class:`~dimod.Vartype.SPIN`, ``'SPIN'``, ``{-1, 1}``
            * :class:`~dimod.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``

            If ``bqm`` is provided, ``vartype`` is not required.

        bqm:
            Terms of the reduced polynomial are added to this binary quadratic model.
            If not provided, a new binary quadratic model is created.

    Examples:

        >>> poly = {(0,): -1, (1,): 1, (2,): 1.5, (0, 1): -1, (0, 1, 2): -2}
        >>> bqm = dimod.make_quadratic(poly, 5.0, dimod.SPIN)

    """
    from dimod.generators import and_gate

    bqm, vartype = _init_quadratic_model(bqm, vartype, BinaryQuadraticModel)
    poly = _init_binary_polynomial(poly, vartype)

    variables = set().union(*poly)
    reduced_terms, constraints = reduce_binary_polynomial(poly)

    for (u, v), p in constraints:

        # add a constraint enforcing the relationship between p == u*v
        if vartype is Vartype.BINARY:
            constraint = and_gate(u, v, p)
            bqm.info['reduction'][(u, v)] = {'product': p}
        elif vartype is Vartype.SPIN:
            aux = _new_aux(variables, u, v)  # need an aux in SPIN-space
            constraint = _spin_product([u, v, p, aux])
            bqm.info['reduction'][(u, v)] = {'product': p, 'auxiliary': aux}
        else:
            raise RuntimeError("unknown vartype: {!r}".format(vartype))

        # scale constraint and update the polynomial with it
        constraint.scale(strength)
        for v, bias in constraint.linear.items():
            bqm.add_variable(v, bias)
        for uv, bias in constraint.quadratic.items():
            bqm.add_interaction(*uv, bias)
        bqm.offset += constraint.offset

    _init_objective(bqm, reduced_terms)

    return bqm


def poly_energy(sample_like: SamplesLike,
                poly: Union[Polynomial, BinaryPolynomial]) -> float:
    """Calculate energy of a sample from a higher order polynomial.

    Args:
         sample_like:
            A raw sample. `samples-like` is an extension of NumPy's
            array_like structure. See :func:`.as_samples`.

        poly (dict):
            Either a polynomial, as a dict of form `{term: bias, ...}`, where `term`
            is a tuple of one or more variables and `bias` the associated bias,
            or a :class:`.BinaryPolynomial`.
            Variable labeling/indexing here must match that of ``sample_like``

    Returns: Energy of the sample.

    Examples:
        >>> poly = dimod.BinaryPolynomial({'a': -1, ('a', 'b'): 1, ('a', 'b', 'c'): -1},
        ...                               dimod.BINARY)
        >>> sample = {'a': 1, 'b': 1, 'c': 0}
        >>> dimod.poly_energy(sample, poly)
        0.0

    """
    return BinaryPolynomial(poly, 'SPIN').energy(sample_like)


def poly_energies(samples_like: SamplesLike,
                  poly: Union[Polynomial, BinaryPolynomial]) -> np.ndarray:
    """Calculates energy of samples from a higher order polynomial.

    Args:
        samples_like:
            A collection of raw samples. `samples-like` is an extension of
            NumPy's array_like structure. See :func:`.as_samples`.

        poly:
            Either a polynomial, as a dict of form `{term: bias, ...}`, where `term`
            is a tuple of one or more variables and `bias` the associated bias,
            or a :class:`.BinaryPolynomial`. Variable labeling/indexing here must
            match that of ``samples_like``.

    Returns: Energies of the samples.

    Examples:
        >>> poly = dimod.BinaryPolynomial({'a': -1, ('a', 'b'): 1, ('a', 'b', 'c'): -1},
        ...                               dimod.BINARY)
        >>> samples = [{'a': 1, 'b': 1, 'c': 0},
        ...            {'a': 1, 'b': 1, 'c': 1}]
        >>> dimod.poly_energies(samples, poly)
        array([ 0., -1.])

    """
    return BinaryPolynomial(poly, 'SPIN').energies(samples_like)



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import typing

import numpy as np

from dimod.binary_quadratic_model import BinaryQuadraticModel
from dimod.constrained import ConstrainedQuadraticModel
from dimod.vartypes import BINARY

__all__ = ['random_bin_packing']


def random_bin_packing(num_items: int,
                       seed: typing.Optional[int] = None,
                       weight_range: typing.Tuple[int, int] = (10, 30),
                       ) -> ConstrainedQuadraticModel:
    """Generate a bin packing problem as a constrained quadratic model.

    The weights for each item are integers uniformly drawn from in the
    ``weight_range``. The bin capacity is set to ``num_items * mean(weights) / 5``.

    Args:
        num_items: Number of items to choose from.

        seed: Seed for NumPy random number generator.

        weight_range: The range of the randomly generated weights for each item.

    Returns:
        The constrained quadratic model encoding the bin packing problem.
        Variables are labeled as ``y_{j}`` where ``y_{j} == 1`` means that bin
        ``j`` has been used and ``x_{i}_{j}`` where ``x_{i}_{j} == 1`` means
        that item ``i`` has been placed in bin ``j``.

    """

    rng = np.random.RandomState(seed)

    max_num_bins = num_items
    weights = list(rng.randint(*weight_range, num_items))
    bin_capacity = int(num_items * np.mean(weights) / 5)
    model = ConstrainedQuadraticModel()

    obj = BinaryQuadraticModel(BINARY)
    y = {j: obj.add_variable(f'y_{j}') for j in range(max_num_bins)}

    for j in range(max_num_bins):
        obj.set_linear(y[j], 1)

    model.set_objective(obj)

    x = {(i, j): model.add_variable(BINARY, f'x_{i}_{j}') for i in range(num_items) for
         j in range(max_num_bins)}

    # Each item goes to one bin
    for i in range(num_items):
        model.add_constraint([(x[(i, j)], 1) for j in range(max_num_bins)] + [(-1,)], sense="==",
                             label='item_placing_{}'.format(i))

    # Bin capacity constraint
    for j in range(max_num_bins):
        model.add_constraint(
            [(x[(i, j)], weights[i]) for i in range(num_items)] + [(y[j], -bin_capacity)],
            sense="<=", label='capacity_bin_{}'.format(j))

    return model



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import abc
import io
import json
import struct
import tempfile
import warnings

from collections import namedtuple
from typing import BinaryIO, ByteString, Callable, Mapping, Tuple

import numpy as np

from dimod.variables import iter_deserialize_variables, iter_serialize_variables


__all__ = ['FileView', 'load']


if issubclass(tempfile.SpooledTemporaryFile, io.IOBase):
    # Python 3.11+
    SpooledTemporaryFile = tempfile.SpooledTemporaryFile
else:
    # we want to use SpooledTemporaryFile but have it also include the methods
    # from io.IOBase.
    class SpooledTemporaryFile(tempfile.SpooledTemporaryFile):
        # This is not part of io.IOBase, but it is implemented in io.BytesIO
        # and io.TextIOWrapper

        def readinto(self, *args, **kwargs):
            return self._file.readinto(*args, **kwargs)

        def readable(self):
            return self._file.readable()

        def seekable(self):
            return self._file.seekable()

        def writable(self):
            return self._file.writable()


class Section(abc.ABC):
    @property
    @abc.abstractmethod
    def magic(self):
        """A 4-byte section identifier. Must be a class variable."""
        pass

    @classmethod
    @abc.abstractmethod
    def loads_data(cls, buff):
        """Accepts a bytes-like object and returns the saved data."""
        pass

    @abc.abstractmethod
    def dump_data(self):
        """Returns a bytes-like object encoding the relevant data."""
        pass

    def dumps(self, **kwargs):
        """Wraps .dump_data to include the identifier and section length."""
        magic = self.magic

        if not isinstance(magic, bytes):
            raise TypeError("magic string should by bytes object")
        if len(magic) != 4:
            raise ValueError("magic string should be 4 bytes in length")

        length = bytes(4)  # placeholder 4 bytes for length

        data = self.dump_data(**kwargs)

        data_length = len(data)

        parts = [magic, length, data]

        if (data_length + len(magic) + len(length)) % 64:
            pad_length = 64 - (data_length + len(magic) + len(length)) % 64
            parts.append(b' '*pad_length)
            data_length += pad_length

        parts[1] = np.dtype('<u4').type(data_length).tobytes()

        assert sum(map(len, parts)) % 64 == 0

        return b''.join(parts)

    @classmethod
    def load(cls, fp, **kwargs):
        """Wraps .loads_data and checks the identifier and length."""
        magic = fp.read(len(cls.magic))
        if magic != cls.magic:
            raise ValueError("unknown subheader, expected {} but recieved "
                             "{}".format(cls.magic, magic))
        length = np.frombuffer(fp.read(4), '<u4')[0]
        return cls.loads_data(fp.read(int(length)), **kwargs)


class VariablesSection(Section):
    magic = b'VARS'

    def __init__(self, variables):
        self.variables = variables

    def dump_data(self):
        serializable = list(iter_serialize_variables(self.variables))
        return json.dumps(serializable).encode('ascii')

    @classmethod
    def loads_data(self, data):
        return iter_deserialize_variables(json.loads(data.decode('ascii')))


def FileView(bqm, version=(1, 0), ignore_labels=False):
    """
    Deprecated. Use :func:`~dimod.binary.BinaryQuadraticModel.to_file` instead.

    .. deprecated:: 0.10.0

        This function will be removed in dimod 0.12.0.

    """
    warnings.warn("FileView is deprecated since dimod 0.10.0 and will be removed in 0.12.0. "
                  "Use `bqm.to_file` instead",
                  DeprecationWarning, stacklevel=2)
    return bqm.to_file(version=version, ignore_labels=ignore_labels)


class _BytesIO(io.RawIOBase):
    # A stub implementation that mimics io.BytesIO but does not make a copy
    # in the case of a memoryview or bytearray. This is necessary because,
    # although io.BytesIO avoids a copy of bytes objects in python 3.5+, it
    # still copies the mutable versions.
    #
    # This is based on the version in the _pyio library
    # https://github.com/python/cpython/blob/3.5/Lib/_pyio.py#L831
    #
    # Copyright 2001-2019 Python Software Foundation; All Rights Reserved
    #
    # 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and
    #    the Individual or Organization ("Licensee") accessing and otherwise using Python
    #    3.5.9 software in source or binary form and its associated documentation.
    #
    # 2. Subject to the terms and conditions of this License Agreement, PSF hereby
    #    grants Licensee a nonexclusive, royalty-free, world-wide license to reproduce,
    #    analyze, test, perform and/or display publicly, prepare derivative works,
    #    distribute, and otherwise use Python 3.5.9 alone or in any derivative
    #    version, provided, however, that PSF's License Agreement and PSF's notice of
    #    copyright, i.e., "Copyright 2001-2019 Python Software Foundation; All Rights
    #    Reserved" are retained in Python 3.5.9 alone or in any derivative version
    #    prepared by Licensee.
    #
    # 3. In the event Licensee prepares a derivative work that is based on or
    #    incorporates Python 3.5.9 or any part thereof, and wants to make the
    #    derivative work available to others as provided herein, then Licensee hereby
    #    agrees to include in any such work a brief summary of the changes made to Python
    #    3.5.9.
    #
    # 4. PSF is making Python 3.5.9 available to Licensee on an "AS IS" basis.
    #    PSF MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.  BY WAY OF
    #    EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR
    #    WARRANTY OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE
    #    USE OF PYTHON 3.5.9 WILL NOT INFRINGE ANY THIRD PARTY RIGHTS.
    #
    # 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON 3.5.9
    #    FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL DAMAGES OR LOSS AS A RESULT OF
    #    MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON 3.5.9, OR ANY DERIVATIVE
    #    THEREOF, EVEN IF ADVISED OF THE POSSIBILITY THEREOF.
    #
    # 6. This License Agreement will automatically terminate upon a material breach of
    #    its terms and conditions.
    #
    # 7. Nothing in this License Agreement shall be deemed to create any relationship
    #    of agency, partnership, or joint venture between PSF and Licensee.  This License
    #    Agreement does not grant permission to use PSF trademarks or trade name in a
    #    trademark sense to endorse or promote products or services of Licensee, or any
    #    third party.
    #
    # 8. By copying, installing or otherwise using Python 3.5.9, Licensee agrees
    #    to be bound by the terms and conditions of this License Agreement.

    def __init__(self, buff):
        self._buffer = memoryview(buff)
        self._pos = 0

    def read(self, size=None):
        if size is None:
            size = -1
        if size < 0:
            size = len(self._buffer)

        if len(self._buffer) <= self._pos:
            return b''
        newpos = min(len(self._buffer), self._pos + size)
        b = self._buffer[self._pos: newpos]
        self._pos = newpos
        return bytes(b)

    def readable(self):
        return True

    def seek(self, pos, whence=0):
        if whence == 0:
            if pos < 0:
                raise ValueError("negative seek position %r" % (pos,))
            self._pos = pos
        elif whence == 1:
            self._pos = max(0, self._pos + pos)
        elif whence == 2:
            self._pos = max(0, len(self._buffer) + pos)
        else:
            raise ValueError("unsupported whence value")
        return self._pos

    def seekable(self):
        return True


_loaders: Mapping[bytes, Callable] = dict()


def register(prefix: bytes, loader: Callable):
    """Register a new loader."""
    _loaders[prefix] = loader


def load(fp, cls=None):
    """Load a model from a file.

    Args:
        fp (bytes-like/file-like):
            If file-like, should be a readable, seekable file-like object. If
            bytes-like, it will be wrapped with :class:`io.BytesIO`.

        cls (class, optional):
            Deprecated keyword argument. Does nothing.

    Returns:
        The loaded model.

    Examples:
        This example saves a binary quadratic model to a
        :func:`~tempfile.TemporaryFile` and then uses :func:`~.load` to read that
        file into a quadratic model.

        >>> from dimod.serialization.fileview import load
        >>> bqm = dimod.BQM({0: 1.0}, {(0, 1): 2, (2, 3): 0.4}, 0.0, dimod.BINARY)
        >>> bqm_file = bqm.to_file()
        >>> _ = bqm_file.seek(0)
        >>> qm = dimod.QuadraticModel.from_bqm(load(bqm_file))

    .. deprecated:: 0.10.0

        The ``cls`` keyword will be removed in dimod 0.12.0. It currently does
        nothing.

    """
    if cls is not None:
        warnings.warn("'cls' keyword argument is deprecated since dimod 0.10.0 and "
                      "will be removed in 0.12.0. It does nothing.",
                      DeprecationWarning, stacklevel=2)

    if isinstance(fp, ByteString):
        file_like: BinaryIO = _BytesIO(fp)  # type: ignore[assignment]
    else:
        file_like = fp

    if not file_like.seekable:
        raise ValueError("expected file-like to be seekable")

    pos = file_like.tell()

    lengths = sorted(set(map(len, _loaders)))
    for num_bytes in lengths:
        prefix = file_like.read(num_bytes)
        file_like.seek(pos)

        try:
            loader = _loaders[prefix]
        except KeyError:
            pass
        else:
            return loader(file_like)

    raise ValueError("cannot load the given file-like")


# for slightly more explicit naming
load.register = register


def make_header(prefix: bytes, data: Mapping, version: Tuple[int, int]) -> bytearray:
    """Construct a header for serializing dimod models.

    The first `len(prefix)` bytes will be exactly the contents of `prefix`.

    The next 1 byte is an unsigned byte: the major version of the file
    format.

    The next 1 byte is an unsigned byte: the minor version of the file
    format.

    The next 4 bytes form a little-endian unsigned int, the length of
    the header data `HEADER_LEN`.

    The next `HEADER_LEN` bytes form the header data. This will be `data`
    json-serialized and encoded with 'ascii'.

    The header is padded with spaces to make the entire length divisible by 64.

    """
    header = bytearray()

    header += prefix
    header += bytes(version)

    version_start = len(header)

    header += bytes(4)  # will hold the data length

    data_start = len(header)

    header += json.dumps(data, sort_keys=True).encode('ascii')
    header += b'\n'

    # want the entire thing to be divisible by 64 for alignment
    if len(header) % 64:
        header += b' '*(64 - len(header) % 64)

    assert not len(header) % 64

    header[version_start:version_start+4] = struct.pack('<I', len(header) - data_start)

    return header


def write_header(file_like: BinaryIO, prefix: bytes, data: Mapping, version: Tuple[int, int]):
    """Write a header constructed by :func:`.make_header` to a file-like."""
    file_like.write(make_header(prefix, data, version))


HeaderInfo = namedtuple('HeaderInfo', ['data', 'version'])


def read_header(file_like: BinaryIO, prefix: bytes) -> HeaderInfo:
    """Read the information from a header constructed by :func:`.make_header`.

    The return value should be accessed by attribute for easy future expansion.
    """
    read_prefix = file_like.read(len(prefix))
    if read_prefix != prefix:
        raise ValueError("unknown file type, expected magic string "
                         f"{prefix!r} but got {read_prefix!r} "
                         "instead")

    version = tuple(file_like.read(2))

    header_len = struct.unpack('<I', file_like.read(4))[0]
    data = json.loads(file_like.read(header_len).decode('ascii'))

    return HeaderInfo(data, version)



# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Un_lt required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import collections.abc
import io
import itertools
import numbers
import typing

from copy import deepcopy

from cython.operator cimport preincrement as inc, dereference as deref
from libc.math cimport ceil, floor
from libcpp.unordered_set cimport unordered_set
from libcpp.utility cimport move
from libcpp.vector cimport vector

import dimod

from dimod.binary import BinaryQuadraticModel
from dimod.constrained.expression import ObjectiveView, ConstraintView
from dimod.cyqmbase cimport cyQMBase
from dimod.cyqmbase.cyqmbase_float64 import BIAS_DTYPE, INDEX_DTYPE
from dimod.cyutilities cimport as_numpy_float
from dimod.cyutilities cimport cppvartype
from dimod.libcpp.abc cimport QuadraticModelBase as cppQuadraticModelBase
from dimod.libcpp.constrained_quadratic_model cimport Sense as cppSense, Penalty as cppPenalty, Constraint as cppConstraint
from dimod.libcpp.vartypes cimport Vartype as cppVartype, vartype_info as cppvartype_info

from dimod.sym import Sense, Eq, Ge, Le
from dimod.variables import Variables
from dimod.vartypes import as_vartype, Vartype
from dimod.views.quadratic import QuadraticViewsMixin


# todo: move to cyutilities?
cdef cppSense cppsense(object sense) except? cppSense.GE:
    if isinstance(sense, str):
        sense = Sense(sense)

    if sense is Sense.Eq:
        return cppSense.EQ
    elif sense is Sense.Le:
        return cppSense.LE
    elif sense is Sense.Ge:
        return cppSense.GE
    else:
        raise RuntimeError(f"unexpected sense: {sense!r}")


cdef class cyConstraintsView:
    cdef cyConstrainedQuadraticModel parent

    def __init__(self, cyConstrainedQuadraticModel parent):
        self.parent = parent

    def __delitem__(self, key) -> None:
        # todo: warn
        raise NotImplementedError

    def __getitem__(self, key) -> ConstraintView:
        cdef Py_ssize_t vi
        try:
            vi = self.parent.constraint_labels.index(key)
        except ValueError as err:
            raise KeyError(repr(key)) from None  # todo: better message

        lhs = ConstraintView(self.parent, key)
        rhs = as_numpy_float(self.parent.cppcqm.constraint_ref(vi).rhs())

        if self.parent.cppcqm.constraint_ref(vi).sense() == cppSense.EQ:
            return Eq(lhs, rhs)
        elif self.parent.cppcqm.constraint_ref(vi).sense() == cppSense.LE:
            return Le(lhs, rhs)
        elif self.parent.cppcqm.constraint_ref(vi).sense() == cppSense.GE:
            return Ge(lhs, rhs)
        else:
            raise RuntimeError("unexpected Sense")

    def __iter__(self):
        yield from self.parent.constraint_labels

    def __len__(self):
        return len(self.parent.constraint_labels)



class ConstraintsView(cyConstraintsView, collections.abc.Mapping):
    def __str__(self) -> str:
        stream = io.StringIO()
        stream.write('{')
        last = len(self) - 1
        for i, (key, value) in enumerate(self.items()):
            stream.write(f'{key!r}: {value!r}')
            if i != last:
                stream.write(', ')
        stream.write('}')
        return stream.getvalue()



cdef class cyConstrainedQuadraticModel:
    def __cinit__(self):
        self.constraint_labels = Variables()
        self.variables = Variables()

        self.dtype = BIAS_DTYPE
        self.index_dtype = INDEX_DTYPE

    def __init__(self):
        self.objective = ObjectiveView(self)

        self.constraints = ConstraintsView(self)

        self.REAL_INTERACTIONS = dimod.REAL_INTERACTIONS

    def __deepcopy__(self, memo):
        cdef cyConstrainedQuadraticModel new = type(self)()

        new.cppcqm = self.cppcqm  # copy assignment operator
        new.constraint_labels = deepcopy(self.constraint_labels, memo)
        new.variables = deepcopy(self.variables, memo)

        memo[id(self)] = new

        return new

    def add_constraint_from_iterable(self, iterable, sense, bias_type rhs, label, weight, penalty):
        # get a fresh constraint        
        constraint = self.cppcqm.new_constraint()

        cdef Py_ssize_t ui, vi
        cdef bias_type bias
        for *variables, bias in iterable:
            if len(variables) == 0:
                constraint.add_offset(bias)
            elif len(variables) == 1:
                vi = self.variables.index(variables[0])  # must already be a variable
                constraint.add_linear(vi, bias)
            elif len(variables) == 2:
                ui = self.variables.index(variables[0])  # must already be a variable
                vi = self.variables.index(variables[1])
                constraint.add_quadratic(ui, vi, bias)
            else:
                raise ValueError("terms must be constant, linear or quadratic")


        constraint.set_sense(cppsense(sense))
        constraint.set_rhs(rhs)

        if weight is not None:
            raise NotImplementedError  # todo

        self.cppcqm.add_constraint(move(constraint))
        label = self.constraint_labels._append(label)

        assert(self.cppcqm.num_constraints() == self.constraint_labels.size())

        return label

    def add_constraint_from_model(self, cyQMBase model, sense, bias_type rhs, label, bint copy, weight, penalty):
        # get a mapping from the model's variables to ours
        cdef vector[Py_ssize_t] mapping
        mapping.reserve(model.num_variables())
        cdef Py_ssize_t vi
        for vi in range(model.num_variables()):
            v = model.variables.at(vi)
            if self.variables.count(v):
                # there is a variable already
                mapping.push_back(self.variables.index(v))

                if self.cppcqm.vartype(mapping[vi]) != model.base.vartype(vi):
                    raise ValueError(f"conflicting vartypes: {v!r}")

                if self.cppcqm.lower_bound(mapping[vi]) != model.base.lower_bound(vi):
                    raise ValueError(f"conflicting lower bounds: {v!r}")

                if self.cppcqm.upper_bound(mapping[vi]) != model.base.upper_bound(vi):
                    raise ValueError(f"conflicting upper bounds: {v!r}")
            else:
                # not yet present, let's just track that fact for now
                # in case there is a mismatch so we don't modify our object yet
                mapping.push_back(-1)
        
        for vi in range(mapping.size()):
            if mapping[vi] != -1:
                continue  # already added and checked

            mapping[vi] = self.cppcqm.num_variables()  # we're about to add a new one

            v = model.variables.at(vi)
            vartype = model.vartype(v)

            self.add_variable(vartype, v,
                              lower_bound=model.base.lower_bound(vi),
                              upper_bound=model.base.upper_bound(vi),
                              )

        cdef Py_ssize_t ci = self.cppcqm.add_constraint(deref(model.base),
                                   cppsense(sense),
                                   rhs,
                                   mapping)

        if weight is not None:
            if penalty == 'linear':
                self.cppcqm.constraint_ref(ci).set_penalty(cppPenalty.LINEAR)
            elif penalty == 'quadratic':
                for vi in range(model.num_variables()):
                    if model.base.vartype(vi) != cppVartype.BINARY and model.base.vartype(vi) != cppVartype.SPIN:
                        raise ValueError("quadratic penalty only allowed if the constraint has binary variables")
                self.cppcqm.constraint_ref(ci).set_penalty(cppPenalty.QUADRATIC)
            elif penalty == 'constant':
                raise NotImplementedError('penalty should be "linear" or "quadratic"')
                # self.cppcqm.constraint_ref(ci).set_penalty(cppPenalty.CONSTANT)
            else:
                raise NotImplementedError('penalty should be "linear" or "quadratic"')

            self.cppcqm.constraint_ref(ci).set_weight(weight)
            

        return self.constraint_labels._append(label)

    def add_variable(self, vartype, v=None, *, lower_bound=None, upper_bound=None):
        cdef cppVartype vt = cppvartype(as_vartype(vartype, extended=True))

        cdef Py_ssize_t vi
        cdef bias_type lb
        cdef bias_type ub

        if v is not None and self.variables.count(v):
            # variable is already present
            vi = self.variables.index(v)
            if self.cppcqm.vartype(vi) != vt:
                raise TypeError(f"variable {v!r} already exists with a different vartype")
            if vt != cppVartype.BINARY and vt != cppVartype.SPIN:
                if lower_bound is not None:
                    lb = lower_bound
                    if lb != self.cppcqm.lower_bound(vi):
                        raise ValueError(
                            f"the specified lower bound, {lower_bound}, for "
                            f"variable {v!r} is different than the existing lower "
                            f"bound, {self.cppcqm.lower_bound(vi)}")
                if upper_bound is not None:
                    ub = upper_bound
                    if ub != self.cppcqm.upper_bound(vi):
                        raise ValueError(
                            f"the specified upper bound, {upper_bound}, for "
                            f"variable {v!r} is different than the existing upper "
                            f"bound, {self.cppcqm.upper_bound(vi)}")

            return v

        # ok, we have a shiny new variable

        if vt == cppVartype.SPIN or vt == cppVartype.BINARY:
            # we can ignore bounds
            v = self.variables._append(v)
            self.cppcqm.add_variable(vt)
            return v

        if vt != cppVartype.INTEGER and vt != cppVartype.REAL:
            raise RuntimeError("unexpected vartype")  # catch some future issues


        
        if lower_bound is None:
            lb = cppvartype_info[bias_type].default_min(vt)
        else:
            lb = lower_bound
            if lb < cppvartype_info[bias_type].min(vt):
                raise ValueError(f"lower_bound cannot be less than {cppvartype_info[bias_type].min(vt)}")

        if upper_bound is None:
            ub = cppvartype_info[bias_type].default_max(vt)
        else:
            ub = upper_bound
            if ub > cppvartype_info[bias_type].max(vt):
                raise ValueError(f"upper_bound cannot be greater than {cppvartype_info[bias_type].max(vt)}")

        v = self.variables._append(v)
        self.cppcqm.add_variable(vt, lb, ub)
        return v

    def change_vartype(self, vartype, v):
        vartype = as_vartype(vartype, extended=True)
        cdef cppVartype vt = cppvartype(vartype)
        cdef Py_ssize_t vi = self.variables.index(v)
        try:
            self.cppcqm.change_vartype(vt, vi)
        except RuntimeError as err:
            # c++ logic_error
            raise TypeError(f"cannot change vartype {self.vartype(v).name!r} "
                            f"to {vartype.name!r}") from None

    def clear(self):
        self.variables._clear()
        self.constraint_labels._clear()
        self.cppcqm.clear()

    def fix_variable(self, v, bias_type assignment):
        self.cppcqm.fix_variable(self.variables.index(v), assignment)
        self.variables._remove(v)

    def flip_variable(self, v):
        cdef Py_ssize_t vi = self.variables.index(v)
        if self.cppcqm.vartype(vi) == cppVartype.SPIN:
            self.cppcqm.substitute_variable(vi, -1, 0)
        elif self.cppcqm.vartype(vi) == cppVartype.BINARY:
            self.cppcqm.substitute_variable(vi, -1, 1)
        else:
            raise ValueError(f"can only flip SPIN and BINARY variables")

    def lower_bound(self, v):
        """Return the lower bound on the specified variable.

        Args:
            v: Variable label for a variable in the model.

        Examples:
            >>> i = dimod.Integer("i", lower_bound=3)
            >>> j = dimod.Integer("j", upper_bound=3)
            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> cqm.add_constraint_from_comparison(i + j >= 4, label="Lower limit")
            'Lower limit'
            >>> cqm.lower_bound("i")
            3.0

        """
        return as_numpy_float(self.cppcqm.lower_bound(self.variables.index(v)))

    def num_constraints(self):
        return self.cppcqm.num_constraints()

    def num_soft_constraints(self):
        cdef Py_ssize_t count = 0
        for c in range(self.cppcqm.num_constraints()):
            if self.cppcqm.constraint_ref(c).is_soft():
                count += 1 
        return count

    def num_variables(self):
        return self.cppcqm.num_variables()

    def remove_constraint(self, label):
        cdef Py_ssize_t ci = self.constraint_labels.index(label)
        self.cppcqm.remove_constraint(ci)
        self.constraint_labels._remove(label)

    def remove_variable(self, v):
        cdef Py_ssize_t vi = self.variables.index(v)
        self.cppcqm.remove_variable(vi)
        self.variables._remove(v)

    def set_lower_bound(self, v, bias_type lb):
        """Set the lower bound for a variable.

        Args:
            v: Variable label of a variable in the constrained quadratic model.
            lb: Lower bound to set for variable ``v``.

        Raises:
            ValueError: If ``v`` is a :class:`~dimod.Vartype.SPIN`
                or :class:`~dimod.Vartype.BINARY` variable.

        Examples:
            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> cqm.add_variable('INTEGER', 'j', upper_bound=5)
            'j'
            >>> cqm.set_lower_bound('j', 2)
        """
        cdef Py_ssize_t vi = self.variables.index(v)

        cdef cppVartype vt = self.cppcqm.vartype(vi)
        if vt == cppVartype.BINARY or vt == cppVartype.SPIN:
            raise ValueError(
                "cannot set the lower bound for BINARY or SPIN variables, "
                f"{v!r} is a {self.vartype(v).name} variable")

        if lb < cppvartype_info[bias_type].min(vt):
            raise ValueError(f"lower_bound cannot be less than {cppvartype_info[bias_type].min(vt)}")

        if lb > self.cppcqm.upper_bound(vi):
            raise ValueError(
                f"the specified lower bound, {lb}, cannot be set greater than the "
                f"current upper bound, {self.cppcqm.upper_bound(vi)}"
                )

        if vt == cppVartype.INTEGER:
            if ceil(lb) > floor(self.cppcqm.upper_bound(vi)):
                raise ValueError(
                    "there must be at least one integer value between "
                    f"the specified lower bound, {lb} and the "
                    f"current upper bound, {self.cppcqm.upper_bound(vi)}"
                    )

        self.cppcqm.set_lower_bound(vi, lb)

    def _set_objective_from_cyqm(self, cyQMBase objective):

        # get a mapping from the objective's variables to ours
        cdef vector[Py_ssize_t] mapping
        mapping.reserve(objective.num_variables())
        cdef Py_ssize_t vi
        for vi in range(objective.num_variables()):
            v = objective.variables.at(vi)
            if self.variables.count(v):
                # there is a variable already
                mapping.push_back(self.variables.index(v))

                if self.cppcqm.vartype(mapping[vi]) != objective.base.vartype(vi):
                    raise ValueError(f"conflicting vartypes: {v!r}")

                if self.cppcqm.lower_bound(mapping[vi]) != objective.base.lower_bound(vi):
                    raise ValueError(f"conflicting lower bounds: {v!r}")

                if self.cppcqm.upper_bound(mapping[vi]) != objective.base.upper_bound(vi):
                    raise ValueError(f"conflicting upper bounds: {v!r}")
            else:
                # not yet present, let's just track that fact for now
                # in case there is a mismatch so we don't modify our object yet
                mapping.push_back(-1)
        
        for vi in range(mapping.size()):
            if mapping[vi] != -1:
                continue  # already added and checked

            mapping[vi] = self.cppcqm.num_variables()  # we're about to add a new one

            v = objective.variables.at(vi)
            vartype = objective.vartype(v)

            self.add_variable(vartype, v,
                              lower_bound=objective.base.lower_bound(vi),
                              upper_bound=objective.base.upper_bound(vi),
                              )

        self.cppcqm.set_objective(deref(objective.base), mapping)

    def set_objective(self, objective):
        """Set the objective of the constrained quadratic model.

        Args:
            objective: Binary quadratic model (BQM) or quadratic model (QM) or
                an iterable of tuples.

        Examples:
            >>> from dimod import Integer, ConstrainedQuadraticModel
            >>> i = Integer('i')
            >>> j = Integer('j')
            >>> cqm = ConstrainedQuadraticModel()
            >>> cqm.set_objective(2*i - 0.5*i*j + 10)

        """
        if isinstance(objective, typing.Iterable):
            terms = objective
        else:
            if isinstance(objective, BinaryQuadraticModel) and objective.dtype == object:
                objective = BinaryQuadraticModel(objective, dtype=self.dtype)

            return self._set_objective_from_cyqm(objective.data)

            # # assume that we're some sort of QM
            # terms = itertools.chain(
            #     objective.iter_linear(),
            #     objective.iter_quadratic(),
            #     ((objective.offset,),)
            #     )

        # clear out anything currently in there. This actually introduces
        # an issue where if the terms later raise an error that we get only
        # a partly formed model. We could mitigate this by dumping the terms
        # to a set of vectors or something, but this avoids that intermediate
        # object
        self.cppcqm.objective.clear()  

        cdef Py_ssize_t ui
        cdef Py_ssize_t vi
        cdef bias_type bias
        for *variables, bias in terms:
            if len(variables) == 0:
                self.cppcqm.objective.add_offset(bias)
            elif len(variables) == 1:
                ui = self.variables.index(variables[0])
                self.cppcqm.objective.add_linear(ui, bias)
            elif len(variables) == 2:
                # quadratic
                ui = self.variables.index(variables[0])
                vi = self.variables.index(variables[1])
                self.cppcqm.objective.add_quadratic(ui, vi, bias)
            else:
                raise ValueError("terms must be a tuple of length 1, 2, or 3")

    def set_upper_bound(self, v, bias_type ub):
        """Set the upper bound for a variable.

        Args:
            v: Variable label of a variable in the constrained quadratic model.
            ub: Upper bound to set for variable ``v``.

        Raises:
            ValueError: If ``v`` is a :class:`~dimod.Vartype.SPIN`
                or :class:`~dimod.Vartype.BINARY` variable.

        Examples:
            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> cqm.add_variable('INTEGER', 'j', lower_bound=2)
            'j'
            >>> cqm.set_upper_bound('j', 5)
        """
        cdef Py_ssize_t vi = self.variables.index(v)
        cdef cppVartype vt = self.cppcqm.vartype(vi)

        if vt == cppVartype.BINARY or vt == cppVartype.SPIN:
            raise ValueError(
                "cannot set the upper bound for BINARY or SPIN variables, "
                f"{v!r} is a {self.vartype(v).name} variable")

        if ub > cppvartype_info[bias_type].max(vt):
            raise ValueError(f"upper_bound cannot be more than {cppvartype_info[bias_type].max(vt)}")
            
        if ub < self.cppcqm.lower_bound(vi):
            raise ValueError(
                f"the specified upper bound, {ub}, cannot be set less than the "
                f"current lower bound, {self.cppcqm.lower_bound(vi)}"
                )

        if vt == cppVartype.INTEGER:
            if ceil(self.cppcqm.lower_bound(vi)) > floor(ub):
                raise ValueError(
                    "there must be at least one integer value between "
                    f"the specified upper bound, {ub} and the "
                    f"current lower bound, {self.cppcqm.lower_bound(vi)}"
                    )

        self.cppcqm.set_upper_bound(vi, ub)

    def upper_bound(self, v):
        """Return the upper bound on the specified variable.

        Args:
            v: Variable label for a variable in the model.

        Examples:
            >>> i = dimod.Integer("i", upper_bound=3)
            >>> j = dimod.Integer("j", upper_bound=3)
            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> cqm.add_constraint_from_comparison(i + j >= 1, label="Upper limit")
            'Upper limit'
            >>> cqm.set_upper_bound("i", 5)
            >>> cqm.upper_bound("i")
            5.0
            >>> cqm.upper_bound("j")
            3.0

        """
        return as_numpy_float(self.cppcqm.upper_bound(self.variables.index(v)))

    def vartype(self, v):
        """Vartype of the given variable.
        
        Args:
            v: Variable label for a variable in the model.
        """
        cdef Py_ssize_t vi = self.variables.index(v)
        cdef cppVartype cppvartype = self.cppcqm.vartype(vi)

        if cppvartype == cppVartype.BINARY:
            return Vartype.BINARY
        elif cppvartype == cppVartype.SPIN:
            return Vartype.SPIN
        elif cppvartype == cppVartype.INTEGER:
            return Vartype.INTEGER
        elif cppvartype == cppVartype.REAL:
            return Vartype.REAL
        else:
            raise RuntimeError("unexpected vartype")


cdef object make_cqm(cppConstrainedQuadraticModel[bias_type, index_type] cppcqm):
    cdef cyConstrainedQuadraticModel cqm = dimod.ConstrainedQuadraticModel()

    cqm.variables._extend(range(cppcqm.num_variables()))
    cqm.constraint_labels._extend(range(cppcqm.num_constraints()))
    cqm.cppcqm = move(cppcqm)

    return cqm



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import math

from dimod.binary.binary_quadratic_model import BinaryQuadraticModel
from dimod.typing import Variable
from dimod.vartypes import Vartype

__all__ = ['binary_encoding']


def binary_encoding(v: Variable, upper_bound: int) -> BinaryQuadraticModel:
    """Generate a binary quadratic model encoding an integer.

    Args:
        v: Integer variable label.
        upper_bound: Upper bound on the integer value (inclusive). You can set
            a lower bound by setting the model's offset.

    Returns:
        A binary quadratic model. Variables in the BQM are labelled with tuples
        with the following two or three values: first, the specified variable
        label, ``v``; second, the coefficient in the integer encoding; third
        (only in the tuple representing the most significant bit), a string,
        ``'msb'``.

    Example:

        >>> bqm = dimod.generators.binary_encoding('i', 6)
        >>> bqm
        BinaryQuadraticModel({('i', 1): 1.0, ('i', 2): 2.0, ('i', 3, 'msb'): 3.0}, {}, 0.0, 'BINARY')

        You can use a sample to restore the original integer value.

        >>> sample = {('i', 1): 1, ('i', 2): 0, ('i', 3, 'msb'): 1}
        >>> bqm.energy(sample)
        4.0
        >>> sum(v[1]*val for v, val in sample.items()) + bqm.offset
        4.0

        If you wish to encode integers with a lower bound, you can use the
        binary quadratic model's :attr:`~dimod.binary.BinaryQuadraticModel.offset`
        attribute.

        >>> i = dimod.generators.binary_encoding('i', 10) + 5  # integer in [5, 15]

    References:
        [1]: Sahar Karimi, Pooya Ronagh (2017), Practical Integer-to-Binary
        Mapping for Quantum Annealers. arxiv.org:1706.01945.

    """
    # note: the paper above also gives a nice way to handle bounded coefficients
    # if we want to do that in the future.

    if upper_bound < 2:
        raise ValueError("upper_bound must be greater than or equal to 2, "
                         f"received {upper_bound}")
    upper_bound = math.floor(upper_bound)

    bqm = BinaryQuadraticModel(Vartype.BINARY)

    max_pow = math.floor(math.log2(upper_bound))
    for exp in range(max_pow):
        val = 1 << exp
        bqm.set_linear((v, val), val)
    else:
        val = upper_bound - ((1 << max_pow) - 1)
        bqm.set_linear((v, val, 'msb'), val)

    return bqm



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Constrained Quadratic Model class.
"""

from __future__ import annotations

import collections.abc
import copy
import io
import json
import os.path
import re
import tempfile
import uuid
import warnings
import zipfile

from io import StringIO
from numbers import Number
from typing import Hashable, Optional, Union, BinaryIO, ByteString, Iterable, Collection, Dict
from typing import Callable, MutableMapping, Iterator, Tuple, Mapping, Any, NamedTuple

import numpy as np

from dimod.binary.binary_quadratic_model import BinaryQuadraticModel, Binary
from dimod.constrained.cyconstrained import cyConstrainedQuadraticModel, ConstraintView, ObjectiveView
from dimod.quadratic.quadratic_model import QuadraticModel
from dimod.sampleset import as_samples
from dimod.serialization.fileview import SpooledTemporaryFile, _BytesIO
from dimod.serialization.fileview import load, read_header, write_header
from dimod.sym import Comparison, Sense
from dimod.typing import Bias, Variable, SamplesLike
from dimod.utilities import new_variable_label
from dimod.variables import serialize_variable, deserialize_variable, Variables
from dimod.vartypes import as_vartype, Vartype, VartypeLike

__all__ = ['ConstrainedQuadraticModel', 'CQM', 'cqm_to_bqm']


CQM_MAGIC_PREFIX = b'DIMODCQM'


class ConstraintData(NamedTuple):
    label: Hashable
    lhs_energy: float
    rhs_energy: float
    sense: Sense
    activity: float
    violation: float


# Previously discrete variables were tracked using `CQM.discrete`. This class
# allows code that made use of that interface to continue to function.
# We should start throwing a deprecation warning in 0.12.1 and remove this in
# 0.14.0
class DiscreteView(collections.abc.MutableSet):
    def __init__(self, parent):
        self.parent = parent

    def __contains__(self, key):
        # warnings.warn("ConstrainedQuadraticModel.discrete attribute is deprecated",
        #               DeprecationWarning, stacklevel=3)
        return key in self.parent.constraint_labels and self.parent.constraints[key].lhs.is_discrete()

    def __iter__(self):
        # warnings.warn("ConstrainedQuadraticModel.discrete attribute is deprecated",
        #               DeprecationWarning, stacklevel=3)
        for lbl, comp in self.parent.constraints.items():
            if comp.lhs.is_discrete():
                yield lbl

    def __len__(self):
        # warnings.warn("ConstrainedQuadraticModel.discrete attribute is deprecated",
        #               DeprecationWarning, stacklevel=3)
        return sum(1 for _ in self)

    def add(self, key):
        # warnings.warn("ConstrainedQuadraticModel.discrete attribute is deprecated",
        #               DeprecationWarning, stacklevel=3)
        self.parent.constraints[key].lhs.mark_discrete(True)

    def difference(self, *others):
        # warnings.warn("ConstrainedQuadraticModel.discrete attribute is deprecated",
        #               DeprecationWarning, stacklevel=3)
        return set(self).difference(*others)

    def discard(self, key):
        # warnings.warn("ConstrainedQuadraticModel.discrete attribute is deprecated",
        #               DeprecationWarning, stacklevel=3)
        if key in self.parent.constraint_labels:
            self.parent.constraints[key].lhs.mark_discrete(False)


class SoftConstraint(NamedTuple):
    weight: float
    penalty: str


# 0.11.6 added `CQM._soft` as a way to track soft constraints. This class
# allows code that made use of that to continue functioning. We should
# start throwing a deprecation warning in 0.12.1 and remove this in 0.13.0.
class SoftView(collections.abc.Mapping):
    def __init__(self, parent):
        self.parent = parent

    def __getitem__(self, key):
        constraint = self.parent.constraints[key].lhs

        if not constraint.is_soft():
            raise KeyError

        return SoftConstraint(constraint.weight(), constraint.penalty())

    def __iter__(self):
        for label, comp in self.parent.constraints.items():
            if comp.lhs.is_soft():
                yield label

    def __len__(self):
        return sum(1 for _ in self)


class ConstrainedQuadraticModel(cyConstrainedQuadraticModel):
    r"""A constrained quadratic model.

    Constrained quadratic models are problems of the form:

    .. math::

        \begin{align}
            \text{Minimize an objective:} & \\
            & \sum_{i} a_i x_i + \sum_{i \le j} b_{ij} x_i x_j + c, \\
            \text{Subject to constraints:} & \\
            & \sum_i a_i^{(m)} x_i + \sum_{i \le j} b_{ij}^{(m)} x_i x_j+ c^{(m)} \circ 0,
            \quad m=1, \dots, M,
        \end{align}

    where :math:`\{ x_i\}_{i=1, \dots, N}` can be binary\ [#]_ or integer
    variables, :math:`a_{i}, b_{ij}, c` are real values,
    :math:`\circ \in \{ \ge, \le, = \}` and  :math:`M` is the total number of constraints.

    .. [#]
        For binary variables, the range of the quadratic-term summation is
        :math:`i < j` because :math:`x^2 = x` for binary values :math:`\{0, 1\}`
        and :math:`s^2 = 1` for spin values :math:`\{-1, 1\}`.

    Constraints are often categorized as either "hard" or "soft". Any hard constraint
    must be satisfied for a solution of the model to qualify as feasible. Soft
    constraints may be violated to achieve an overall good solution. By setting
    appropriate weights to soft constraints in comparison to the objective
    and to other soft constraints, you can express the relative importance of such
    constraints.

    The objective and constraints are encoded as either :class:`.QuadraticModel`
    or :class:`.BinaryQuadraticModel` depending on the variable types used.

    Example:

        This example solves the simple `bin packing problem <https://w.wiki/3jz4>`_
        of packing a set of items of different weights into the smallest
        possible number of bins.

        `dimod` provides a general :func:`~dimod.generators.random_bin_packing`
        function to generate bin packing problems, and this example follows the
        same naming conventions.

        Consider four objects with weights between 0 and 1, and assume that each
        bin has a capacity to hold up to a total weight of 1.

        >>> weights = [.9, .7, .2, .1]
        >>> capacity = 1

        Variable :math:`y_j` indicates that bin :math:`j` is used. Clearly, no
        more than four bins are needed.

        >>> y = [dimod.Binary(f'y_{j}') for j in range(len(weights))]

        Variable :math:`x_{i,j}` indicates that item :math:`i` is put in bin
        :math:`j`.

        >>> x = [[dimod.Binary(f'x_{i}_{j}') for j in range(len(weights))]
        ...      for i in range(len(weights))]

        Create an empty constrained quadratic model ("empty" meaning that no
        objective or constraints have set).

        >>> cqm = dimod.ConstrainedQuadraticModel()

        The problem is to minimize the number of bins used. Therefore the objective
        is to minimize the value of :math:`\sum_j y_j`.

        >>> cqm.set_objective(sum(y))

        Any feasible solution must meet the constraint that each item can only go
        in one bin. You can express this constraint, for a given item :math:`i`,
        with :math:`\sum_j x_{i, j} == 1`. Note that the label of each
        constraint is returned so that you can access them in the future if
        desired.

        >>> for i in range(len(weights)):
        ...     cqm.add_constraint(sum(x[i]) == 1, label=f'item_placing_{i}')
        'item_placing_0'
        'item_placing_1'
        'item_placing_2'
        'item_placing_3'

        Finally, enforce the limits on each bin. You can express this constraint,
        for a given bin :math:`j`, with :math:`\sum_i x_{i, j} * w_i <= c` where
        :math:`w_i` is the weight of item :math:`i` and :math:`c` is the capacity.

        >>> for j in range(len(weights)):
        ...     cqm.add_constraint(
        ...         sum(weights[i] * x[i][j] for i in range(len(weights))) - y[j] * capacity <= 0,
        ...         label=f'capacity_bin_{j}')
        'capacity_bin_0'
        'capacity_bin_1'
        'capacity_bin_2'
        'capacity_bin_3'

    """
    def __init__(self):
        super().__init__()

        self.discrete: Set[Hashable] = DiscreteView(self)
        self._soft = SoftView(self)  # todo: remove

    def add_constraint(self, data, *args, **kwargs) -> Hashable:
        """Add a constraint to the model.

        This method dispatches to one of several specific methods based on
        the type of the first argument.
        For a detailed description of the accepted argument types, see
        :meth:`~.ConstrainedQuadraticModel.add_constraint_from_model`,
        :meth:`~.ConstrainedQuadraticModel.add_constraint_from_comparison`,
        and :meth:`~.ConstrainedQuadraticModel.add_constraint_from_iterable`.

        """
        # in python 3.8+ we can use singledispatchmethod
        if isinstance(data, (BinaryQuadraticModel, QuadraticModel)):
            return self.add_constraint_from_model(data, *args, **kwargs)
        elif isinstance(data, Comparison):
            return self.add_constraint_from_comparison(data, *args, **kwargs)
        elif isinstance(data, Iterable):
            return self.add_constraint_from_iterable(data, *args, **kwargs)
        else:
            raise TypeError("unexpected data format")

    def add_constraint_from_model(self,
                                  qm: Union[BinaryQuadraticModel, QuadraticModel],
                                  sense: Union[Sense, str],
                                  rhs: Bias = 0,
                                  label: Optional[Hashable] = None,
                                  *,
                                  copy: bool = True,
                                  weight: Optional[float] = None,
                                  penalty: str = 'linear',
                                  ) -> Hashable:
        """Add a constraint from a quadratic model.

        Args:
            qm: Quadratic model or binary quadratic model.

            sense: One of `<=`, `>=`, `==`.

            rhs: Right-hand side of the constraint.

            label: Label for the constraint. Must be unique. If no label
                is provided, then one is generated using :mod:`uuid`.

            copy: If `True`, model ``qm`` is copied. This can be set to `False`
                to improve performance, but subsequently mutating ``qm`` can
                cause issues.

            weight: Weight for a soft constraint. If ``None``, the constraint
                is hard. In feasible solutions, all the model's hard constraints
                must be met, while soft constraints might be violated to achieve
                overall good solutions.

            penalty: Penalty type for a soft constraint (a constraint with its
                ``weight`` parameter set). Supported values are ``'linear'`` and
                ``'quadratic'``. Ignored if ``weight`` is ``None``. ``'quadratic'``
                is supported for a constraint with binary variables only.

        Returns:
            Label of the added constraint.

        Examples:
            This example adds a constraint from the single-variable binary
            quadratic model ``x``.

            >>> from dimod import ConstrainedQuadraticModel, Binary
            >>> cqm = ConstrainedQuadraticModel()
            >>> x = Binary('x')
            >>> cqm.add_constraint_from_model(x, '>=', 0, 'Min x')
            'Min x'
            >>> print(cqm.constraints["Min x"].to_polystring())
            x >= 0.0

        """
        if label is None:
            label = self._new_constraint_label()
        elif label in self.constraint_labels:
            raise ValueError("a constraint with that label already exists")

        if isinstance(qm, BinaryQuadraticModel) and qm.dtype == object:
            qm = BinaryQuadraticModel(qm)

        return super().add_constraint_from_model(
            qm.data,
            sense,
            rhs,
            label,
            bool(copy),
            weight,
            penalty,
            )

    def add_constraint_from_comparison(self,
                                       comp: Comparison,
                                       label: Optional[Hashable] = None,
                                       copy: bool = True,
                                       weight: Optional[float] = None,
                                       penalty: str = 'linear',
                                       ) -> Hashable:
        r"""Add a constraint from a symbolic comparison.

        For a more detailed discussion of symbolic model manipulation, see
        :ref:`intro_symbolic_math`.

        Args:
            comp: Comparison object, generally constructed using symbolic math.
                The right-hand side of any symbolic equation must be an integer
                or float.

            label: Label for the constraint. Must be unique. If no label
                is provided, one is generated using :mod:`uuid`.

            copy: If `True`, the model used in the comparison is copied. You can
                set to `False` to improve performance, but subsequently mutating
                the model can cause issues.

            weight: Weight for a soft constraint. If ``None``, the constraint
                is hard. In feasible solutions, all the model's hard constraints
                must be met, while soft constraints might be violated to achieve
                overall good solutions.

            penalty: Penalty type for a soft constraint (a constraint with its
                ``weight`` parameter set). Supported values are ``'linear'`` and
                ``'quadratic'``. Ignored if ``weight`` is ``None``. ``'quadratic'``
                is supported for a constraint with binary variables only.

        Returns:
            Label of the added constraint.

        Example:

            Encode a constraint.

            .. math::

                x + y + xy <= 1

            First create the relevant variables and the model.

            >>> x, y = dimod.Binaries(['x', 'y'])
            >>> cqm = dimod.ConstrainedQuadraticModel()

            And add the constraint to the model.

            >>> cqm.add_constraint(x + y + x*y <= 1, label='c0')
            'c0'
            >>> cqm.constraints['c0'].to_polystring()
            'x + y + x*y <= 1.0'

        Example:

            Encode a constraint with a symbolic right-hand side.


            .. math::

                x + y \le x y


            First create the relevant variables and the model.

            >>> x, y = dimod.Binaries(['x', 'y'])
            >>> cqm = dimod.ConstrainedQuadraticModel()

            Trying to directly compare the left-hand and right-hand side
            of the equation will raise an error, because the right-hand side
            is not an integer or float.

            >>> try:
            ...     cqm.add_constraint(x + y <= x * y)
            ... except TypeError:
            ...     print("Not allowed!")
            Not allowed!

            To avoid this, simply subtract the right-hand side from both sides.

            .. math::

                x + y - xy \le 0

            The constraint can then be added.

            >>> cqm.add_constraint(x + y - x*y <= 0, label="c0")
            'c0'

        """
        if not isinstance(comp.rhs, Number):
            raise TypeError("comparison should have a numeric rhs")

        if isinstance(comp.lhs, (BinaryQuadraticModel, QuadraticModel)):
            return self.add_constraint_from_model(comp.lhs, comp.sense, rhs=comp.rhs,
                                                  label=label, copy=copy, weight=weight,
                                                  penalty=penalty)
        else:
            raise ValueError("comparison should have a binary quadratic model "
                             "or quadratic model lhs.")

    def add_constraint_from_iterable(self, iterable: Iterable,
                                     sense: Union[Sense, str],
                                     rhs: Bias = 0,
                                     label: Optional[Hashable] = None,
                                     weight: Optional[float] = None,
                                     penalty: str = 'linear',
                                     ) -> Hashable:
        """Add a constraint from an iterable of tuples.

        Args:
            iterable: Iterable of terms as tuples. The variables must
                have already been added to the object.

            sense: One of `<=`, `>=`, `==`.

            rhs: The right-hand side of the constraint.

            label: Label for the constraint. Must be unique. If no label
                is provided, then one is generated using :mod:`uuid`.

            weight: Weight for a soft constraint. If ``None``, the constraint
                is hard. In feasible solutions, all the model's hard constraints
                must be met, while soft constraints might be violated to achieve
                overall good solutions.

            penalty: Penalty type for a soft constraint (a constraint with its
                ``weight`` parameter set). Supported values are ``'linear'`` and
                ``'quadratic'``. Ignored if ``weight`` is ``None``. ``'quadratic'``
                is supported for a constraint with binary variables only.

        Returns:
            Label of the added constraint.

        Examples:
            >>> from dimod import ConstrainedQuadraticModel, Integer, Binary
            >>> cqm = ConstrainedQuadraticModel()
            >>> cqm.add_variable('INTEGER', 'i')
            'i'
            >>> cqm.add_variable('INTEGER', 'j')
            'j'
            >>> cqm.add_variable('BINARY', 'x')
            'x'
            >>> cqm.add_variable('BINARY', 'y')
            'y'
            >>> label1 = cqm.add_constraint_from_iterable([('x', 'y', 1), ('i', 2), ('j', 3),
            ...                                           ('i', 'j', 1)], '<=', rhs=1)
            >>> print(cqm.constraints[label1].to_polystring())
            2*i + 3*j + y*x + i*j <= 1.0

        """
        if label is None:
            label = self._new_constraint_label()
        elif label in self.constraint_labels:
            raise ValueError("a constraint with that label already exists")

        return super().add_constraint_from_iterable(iterable, sense, rhs, label, weight, penalty)

    def add_discrete(self, data, *args, **kwargs) -> Hashable:
        """A convenience wrapper for other methods that add one-hot constraints.

        One-hot constraints can represent discrete variables (for example a
        ``color`` variable that has values ``{"red", "blue", "green"}``) by
        requiring that only one of a set of two or more binary variables is
        assigned a value of 1.

        These constraints support only :class:`~dimod.Vartype.BINARY` variables
        and must be disjoint; that is, variables in one such constraint must not
        be used in others in the model.

        Constraints added by the methods wrapped by :meth:`add_discrete` are
        guaranteed to be satisfied in solutions returned
        by the :class:`~dwave.system.samplers.LeapHybridCQMSampler` hybrid sampler.

        See also:
            :meth:`~.ConstrainedQuadraticModel.add_discrete_from_model`

            :meth:`~.ConstrainedQuadraticModel.add_discrete_from_comparison`

            :meth:`~.ConstrainedQuadraticModel.add_discrete_from_iterable`

        Examples:

            >>> cqm = dimod.ConstrainedQuadraticModel()

            Add a discrete constraint over variables ``x, y, z`` from an
            iterable.

            >>> iterable = ['x', 'y', 'z']
            >>> for v in iterable:
            ...      cqm.add_variable('BINARY', v)
            'x'
            'y'
            'z'
            >>> cqm.add_discrete(iterable, label='discrete-xyz')
            'discrete-xyz'

            Add a discrete constraint over variables ``a, b, c`` from a
            model.

            >>> a, b, c = dimod.Binaries(['a', 'b', 'c'])
            >>> cqm.add_discrete(sum([a, b, c]), label='discrete-abc')
            'discrete-abc'

            Add a discrete constraint over variables ``d, e, f`` from a
            comparison.

            >>> d, e, f = dimod.Binaries(['d', 'e', 'f'])
            >>> cqm.add_discrete(d + e + f == 1, label='discrete-def')
            'discrete-def'

        """
        # in python 3.8+ we can use singledispatchmethod
        if isinstance(data, (BinaryQuadraticModel, QuadraticModel)):
            return self.add_discrete_from_model(data, *args, **kwargs)
        elif isinstance(data, Comparison):
            return self.add_discrete_from_comparison(data, *args, **kwargs)
        elif isinstance(data, Iterable):
            return self.add_discrete_from_iterable(data, *args, **kwargs)
        else:
            raise TypeError("unexpected data format")

    def add_discrete_from_comparison(self,
                                     comp: Comparison,
                                     label: Optional[Hashable] = None,
                                     copy: bool = True) -> Hashable:
        """Add a one-hot constraint from a comparison.

        One-hot constraints can represent discrete variables (for example a
        ``color`` variable that has values ``{"red", "blue", "green"}``) by
        requiring that only one of a set of two or more binary variables is
        assigned a value of 1.

        These constraints support only :class:`~dimod.Vartype.BINARY` variables
        and must be disjoint; that is, variables in such a constraint must not be
        used elsewhere in the model.

        Constraints added by this method are guaranteed to be satisfied in
        solutions returned by the :class:`~dwave.system.samplers.LeapHybridCQMSampler`
        hybrid sampler.

        Args:
            comp: Comparison object. The comparison must be a linear
                equality constraint with all of the linear biases on the
                left-hand side equal to one and the right-hand side equal
                to one.

            label: Label for the constraint. Must be unique. If no label
                is provided, one is generated using :mod:`uuid`.

            copy: If `True`, the model used in the comparison is copied. You can
                set to `False` to improve performance, but subsequently mutating
                the model can cause issues.

        Returns:
            Label of the added constraint.

        Examples:

            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> r, b, g = dimod.Binaries(["red", "blue", "green"])
            >>> cqm.add_discrete_from_comparison(r + b + g == 1, label="One color")
            'One color'

        """
        if comp.sense is not Sense.Eq:
            raise ValueError("discrete constraints must be equality constraints")
        if comp.rhs != 1:
            # could scale, but let's keep it simple for now
            raise ValueError("the right-hand side of a discrete constraint must be 1")
        return self.add_discrete_from_model(comp.lhs, label=label, copy=copy)

    def add_discrete_from_iterable(self,
                                   variables: Iterable[Variable],
                                   label: Optional[Hashable] = None) -> Hashable:
        """Add a one-hot constraint from an iterable.

        One-hot constraints can represent discrete variables (for example a
        ``color`` variable that has values ``{"red", "blue", "green"}``) by
        requiring that only one of a set of two or more binary variables is
        assigned a value of 1.

        These constraints support only :class:`~dimod.Vartype.BINARY` variables
        and must be disjoint; that is, variables in such a constraint must not be
        used elsewhere in the model.

        Constraints added by this method are guaranteed to be satisfied in
        solutions returned by the :class:`~dwave.system.samplers.LeapHybridCQMSampler`
        hybrid sampler.

        Args:
            variables: An iterable of variables.

            label: Label for the constraint. Must be unique. If no label
                is provided, then one is generated using :mod:`uuid`.

        Returns:
            Label of the added constraint.

        Examples:

            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> color = ["red", "blue", "green"]
            >>> for v in color:
            ...      cqm.add_variable('BINARY', v)
            'red'
            'blue'
            'green'
            >>> cqm.add_discrete(color, label='one-color')
            'one-color'
            >>> print(cqm.constraints['one-color'].to_polystring())
            red + blue + green == 1.0

        """
        if label is not None and label in self.constraints:
            raise ValueError("a constraint with that label already exists")

        bqm = BinaryQuadraticModel(Vartype.BINARY, dtype=np.float32)

        for v in variables:
            if v in self.variables:
                # it already exists, let's make sure it's not already used
                if any(v in self.constraints[label].lhs.variables for label in self.discrete):
                    raise ValueError(f"variable {v!r} is already used in a discrete variable")
                if self.vartype(v) is not Vartype.BINARY:
                    raise ValueError(f"variable {v!r} has already been added but is not BINARY")

            bqm.set_linear(v, 1)

        if bqm.num_variables < 2:
            raise ValueError("discrete constraints must have at least two variables")

        label = self.add_constraint_from_comparison(bqm == 1, label=label, copy=False)
        self.discrete.add(label)
        return label

    def add_discrete_from_model(self,
                                qm: Union[BinaryQuadraticModel, QuadraticModel],
                                label: Optional[Hashable] = None,
                                copy: bool = True) -> Hashable:
        """Add a one-hot constraint from a model.

        One-hot constraints can represent discrete variables (for example a
        ``color`` variable that has values ``{"red", "blue", "green"}``) by
        requiring that only one of a set of two or more binary variables is
        assigned a value of 1.

        These constraints support only :class:`~dimod.Vartype.BINARY` variables
        and must be disjoint; that is, variables in such a constraint must not be
        used elsewhere in the model.

        Constraints added by this method are guaranteed to be satisfied in
        solutions returned by the :class:`~dwave.system.samplers.LeapHybridCQMSampler`
        hybrid sampler.

        Args:
            qm: A quadratic model or binary quadratic model.
                The model must be linear with all of the linear biases on the
                left-hand side equal to one and the right-hand side equal
                to one.

            label: A label for the constraint. Must be unique. If no label
                is provided, one is generated using :mod:`uuid`.

            copy: If `True`, the model is copied. You can set to `False` to
                improve performance, but subsequently mutating the model can
                cause issues.

        Returns:
            Label of the added constraint.

        Examples:

            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> r, b, g = dimod.Binaries(["red", "blue", "green"])
            >>> cqm.add_discrete(sum([r, g, b]), label="One color")
            'One color'
            >>> print(cqm.constraints["One color"].to_polystring())
            red + green + blue == 1.0

        """
        vartype = qm.vartype if isinstance(qm, QuadraticModel) else lambda v: qm.vartype

        if qm.num_interactions:
            raise ValueError("discrete constraints must be linear")

        if qm.num_variables < 2:
            raise ValueError("discrete constraints must have at least two variables")

        for v, bias in qm.iter_linear():
            if v in self.variables:
                # it already exists, let's make sure it's not already used
                if any(v in self.constraints[label].lhs.variables for label in self.discrete):
                    raise ValueError(f"variable {v!r} is already used in a discrete variable")
                if self.vartype(v) is not Vartype.BINARY:
                    raise ValueError(f"variable {v!r} has already been added but is not BINARY")
            elif not vartype(v) is Vartype.BINARY:
                raise ValueError("all variables in a discrete constraint must be binary, "
                                 f"{v!r} is {vartype(v).name!r}")
            # we could maybe do a scaling, but let's just keep it simple for now
            if bias != 1:
                raise ValueError("all linear biases in a discrete constraint must be 1")

        label = self.add_constraint_from_model(qm, sense='==', rhs=1, label=label, copy=copy)
        self.discrete.add(label)
        return label

    def add_variable(self, vartype: VartypeLike, v: Optional[Variable] = None,
                     *,
                     lower_bound: Optional[float] = None,
                     upper_bound: Optional[float] = None,
                     ) -> Variable:
        """Add a variable to the model.

        Args:
            vartype:
                Variable type. One of:

                * :class:`~dimod.Vartype.SPIN`, ``'SPIN'``, ``{-1, 1}``
                * :class:`~dimod.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``
                * :class:`~dimod.Vartype.INTEGER`, ``'INTEGER'``
                * :class:`~dimod.Vartype.REAL`, ``'REAL'``

            v:
                Label for the variable. Defaults to the length of the
                quadratic model, if that label is available. Otherwise defaults
                to the lowest available positive integer label.

            lower_bound:
                Lower bound on the variable. Ignored when the variable is
                :class:`~dimod.Vartype.BINARY` or :class:`~dimod.Vartype.SPIN`.

            upper_bound:
                Upper bound on the variable. Ignored when the variable is
                :class:`~dimod.Vartype.BINARY` or :class:`~dimod.Vartype.SPIN`.

        Returns:
            Variable label.

        Raises:
            ValueError: If ``v`` is already a variable in the model and
                the ``vartype``, ``lower_bound``, or ``upper_bound`` are
                inconsistent.

        Examples:
            >>> from dimod import ConstrainedQuadraticModel, Integer
            >>> cqm = ConstrainedQuadraticModel()
            >>> cqm.add_variable('INTEGER', 'i')
            'i'

        """
        try:
            vartype = as_vartype(vartype, extended=True)
        except TypeError:
            # in dimod<0.11 the argument order was v, vartype so let's allow that case
            warnings.warn(
                "Parameter order CQM.add_variable(v, vartype) "
                "is deprecated since dimod 0.11.0 and will be removed in 0.13.0. "
                "Use CQM.add_variable(vartype, v) instead.",
                DeprecationWarning, stacklevel=2)
            v, vartype = vartype, v

        return super().add_variable(vartype, v, lower_bound=lower_bound, upper_bound=upper_bound)

    def add_variables(self,
                      vartype: dimod.VartypeLike,
                      variables: Union[int, Sequence[Variable]],
                      *,
                      lower_bound: Optional[float] = None,
                      upper_bound: Optional[float] = None,
                      ) -> None:
        if isinstance(variables, Iterable):
            for v in variables:
                self.add_variable(vartype, v, lower_bound=lower_bound, upper_bound=upper_bound)
        else:
            for _ in range(variables):
                self.add_variable(vartype, lower_bound=lower_bound, upper_bound=upper_bound)

    def check_feasible(self, sample_like: SamplesLike, rtol: float = 1e-6, atol: float = 1e-8) -> bool:
        r"""Return the feasibility of the given sample.

        A sample is feasible if all constraints are satisfied. A constraint's
        satisfaction is tested using the following equation:

        .. math::

            violation <= (atol + rtol * | rhs\_energy | )

        where ``violation`` and ``rhs_energy`` are as returned by :meth:`.iter_constraint_data`.

        Args:
            sample_like: A sample. `sample-like` is an extension of
                NumPy's array_like structure. See :func:`.as_samples`.
            rtol: Relative tolerance.
            atol: Absolute tolerance.

        Returns:
            True if the sample is feasible (given the tolerances).

        Examples:
            This example violates a constraint that :math:`i \le 4` by `0.2`,
            which is greater than the absolute tolerance set by ``atol = 0.1``
            but within the relative tolerance of :math:`0.1 * 4 = 0.4`.

            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> i = dimod.Integer("i")
            >>> cqm.add_constraint_from_comparison(i <= 4, label="Max i")
            'Max i'
            >>> cqm.check_feasible({"i": 4.2}, atol=0.1)
            False
            >>> next(cqm.iter_constraint_data({"i": 4.2})).rhs_energy
            4.0
            >>> cqm.check_feasible({"i": 4.2}, rtol=0.1)
            True

            Note that the :func:`next` function is used here because the model
            has just a single constraint.
        """
        return all(datum.violation <= atol + rtol*abs(datum.rhs_energy)
                   for datum in self.iter_constraint_data(sample_like))

    def fix_variable(self, v: Variable, value: float, *,
                     cascade: Optional[bool] = None,
                     ) -> Dict[Variable, float]:
        """Fix the value of a variable in the model.

        Note that this function does not test feasibility.

        Args:
            v: Variable label for a variable in the model.

            value: Value to assign variable ``v``.

            cascade: Deprecated. Does nothing.

        Returns:
            An empty dictionary, for legacy reasons.

        Raises:
            ValueError: If ``v`` is not the label of a variable in the model.

        .. deprecated:: 0.12.0
            The ``cascade`` keyword argument will be removed in 0.14.0.
            It currently does nothing.

        """
        if cascade is not None:
            warnings.warn("The 'cascade' keyword argument is deprecated since dimod 0.12.0 "
                          "and will be removed in 0.14.0", DeprecationWarning,
                          stacklevel=2)

        # get the discrete constaints that are possibly affected
        discrete = dict()
        for label in self.discrete:
            lhs = self.constraints[label].lhs
            if v in lhs.variables:
                discrete[label] = lhs

        super().fix_variable(v, value)

        if discrete:
            if value == 1:
                self.discrete -= discrete
            else:
                for label, lhs in discrete.items():
                    if lhs.num_variables <= 1:
                        self.discrete.discard(label)
        return {}

    def fix_variables(self,
                      fixed: Union[Mapping[Variable, float],
                                          Iterable[Tuple[Variable, float]]],
                      *,
                      cascade: Optional[bool] = None,
                      ) -> Dict[Variable, float]:
        """Fix the value of the variables and remove them.

        Args:
            fixed: Dictionary or iterable of 2-tuples of variable assignments.
            cascade: Deprecated. Does nothing.

        Returns:
            An empty dictionary, for legacy reasons.

        .. deprecated:: 0.12.0
            The ``cascade`` keyword argument will be removed in 0.14.0.
            It currently does nothing.

        """
        if isinstance(fixed, Mapping):
            fixed = fixed.items()

        for v, val in fixed:
            self.fix_variable(v, val, cascade=cascade)

        return {}

    def flip_variable(self, v: Variable):
        r"""Flip the specified binary variable in the objective and constraints.

        Note that this may terminate a constraint's status as a discrete constraint
        (see :meth:`add_discrete`). Subsequently flipping the variable again does
        not restore that status.

        Args:
            v: Variable label of a :class:`~dimod.Vartype.BINARY` or
                :class:`~dimod.Vartype.SPIN` variable.

        Raises:
            ValueError: If given a non-binary variable to flip.

        Examples:
            This example flips :math:`x`` in an objective, :math:`2xy-2x`, which
            is equivalent for binary variables with values :math:`\{0, 1\}` to
            the substitution :math:`x \Rightarrow 1-x`, creating a new objective,
            :math:`2xy-2x \Rightarrow 2(1-x)y -2(1-x) = 2y -2xy -2 -2x`.

            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> x, y = dimod.Binaries(["x", "y"])
            >>> cqm.set_objective(2 * x * y - 2 * x)
            >>> cqm.flip_variable("x")
            >>> print(cqm.objective.to_polystring())
            -2 + 2*x + 2*y - 2*x*y

            The next example flips a variable in a one-hot constraint. Subsequently
            fixing one of this discrete constaint's binary variables to `1`
            does not ensure the remaining variables are assigned `0`.

            >>> r, b, g = dimod.Binaries(["red", "blue", "green"])
            >>> cqm.add_discrete_from_comparison(r + b + g == 1, label="One color")
            'One color'
            >>> cqm.flip_variable("red")
            >>> cqm.fix_variable("blue", 1, cascade=True)
            {}

        """
        super().flip_variable(v)

        for label in list(self.discrete):
            lhs = self.constraints[label].lhs
            if v in lhs.variables:
                self.discrete.discard(label)  # no longer a discrete variable

    @classmethod
    def from_bqm(cls, bqm: BinaryQuadraticModel) -> ConstrainedQuadraticModel:
        """Alias for :meth:`from_quadratic_model`."""
        return cls.from_quadratic_model(bqm)

    @classmethod
    def from_discrete_quadratic_model(cls, dqm: DiscreteQuadraticModel, *,
                                      relabel_func: Callable[[Variable, int], Variable] = lambda v, c: (v, c),
                                      ) -> ConstrainedQuadraticModel:
        """Construct a constrained quadratic model from a discrete quadratic model.

        Args:
            dqm: Discrete quadratic model.

            relabel_func (optional): A function that takes two arguments, the
                variable label and the case label, and returns a new variable
                label to be used in the CQM. By default generates a 2-tuple
                `(variable, case)`.

        Returns:
            A constrained quadratic model.

        """
        cqm = cls()

        objective = BinaryQuadraticModel(Vartype.BINARY)

        seen = set()
        for v in dqm.variables:
            seen.add(v)

            # convert v, case to a flat set of variables
            v_vars = list(relabel_func(v, case) for case in dqm.get_cases(v))

            # add the one-hot constraint
            cqm.add_discrete(v_vars, label=v)

            # add to the objective
            objective.add_linear_from(zip(v_vars, dqm.get_linear(v)))

            for u in dqm.adj[v]:
                if u in seen:  # only want upper-triangle
                    continue

                u_vars = list(relabel_func(u, case) for case in dqm.get_cases(u))

                objective.add_quadratic_from(
                    (u_vars[cu], v_vars[cv], bias)
                    for (cu, cv), bias
                    in dqm.get_quadratic(u, v).items()
                    )

        objective.offset = dqm.offset

        cqm.set_objective(objective)

        return cqm

    from_dqm = from_discrete_quadratic_model

    @classmethod
    def from_qm(cls, qm: QuadraticModel) -> ConstrainedQuadraticModel:
        """Alias for :meth:`from_quadratic_model`."""
        return cls.from_quadratic_model(qm)

    @classmethod
    def from_quadratic_model(cls, qm: Union[QuadraticModel, BinaryQuadraticModel]
                             ) -> ConstrainedQuadraticModel:
        """Construct a constrained quadratic model from a quadratic model or
        binary quadratic model.

        The specified model is set as the objective to be minimzed in the constructed
        constrained quadratic model (CQM). You can then add constraints that
        feasible solutions should meet.

        Args:
            qm: Binary quadratic model (BQM) or quadratic model (QM).

        Examples:
            This example creates a CQM to minimize a triangular problem with the added
            constraint that one of the variables must have value 1 in feasible solutions.

            >>> from dimod import ConstrainedQuadraticModel, BinaryQuadraticModel
            >>> bqm = BinaryQuadraticModel.from_ising({}, {'ab': 1, 'bc': 1, 'ac': 1})
            >>> cqm = ConstrainedQuadraticModel().from_bqm(bqm)
            >>> cqm.objective.linear
            {'a': 0.0, 'b': 0.0, 'c': 0.0}
            >>> cqm.objective.quadratic
            {('b', 'a'): 1.0, ('c', 'a'): 1.0, ('c', 'b'): 1.0}
            >>> label1 = cqm.add_constraint_from_model(BinaryQuadraticModel({'a': 0}, {}, 0, 'SPIN'), '>=', 0)
        """
        cqm = cls()
        cqm.set_objective(qm)
        return cqm

    @classmethod
    def from_file(cls,
                  fp: Union[BinaryIO, ByteString],
                  *,
                  check_header: bool = True,
                  ) -> ConstrainedQuadraticModel:
        """Construct from a file-like object.

        Args:
            fp: File pointer to a readable, seekable file-like object.

            check_header: If True, the header is checked for consistency
                against the deserialized model. Otherwise it is ignored.

        The inverse of :meth:`~ConstrainedQuadraticModel.to_file`.

        Examples:
            >>> cqm1 = dimod.ConstrainedQuadraticModel()
            >>> x, y = dimod.Binaries(["x", "y"])
            >>> cqm1.set_objective(2 * x * y - 2 * x)
            >>> cqm_file = cqm1.to_file()
            >>> cqm2 = dimod.ConstrainedQuadraticModel.from_file(cqm_file)
            >>> print(cqm2.objective.to_polystring())
            -2*x + 2*x*y

        """
        if isinstance(fp, ByteString):
            file_like: BinaryIO = _BytesIO(fp)  # type: ignore[assignment]
        else:
            file_like = fp

        header_info = read_header(file_like, CQM_MAGIC_PREFIX)

        if not (1, 0) <= header_info.version <= (1, 3):
            raise ValueError("cannot load a CQM serialized with version "
                             f"{header_info.version!r}, try upgrading your "
                             "dimod version")

        # we don't actually need the data

        cqm = CQM()

        with zipfile.ZipFile(file_like, mode='r') as zf:
            cqm.set_objective(load(zf.read("objective")))

            constraint_labels = set()
            for arch in zf.namelist():
                # even on windows zip uses /
                match = re.match("constraints/([^/]+)/", arch)
                if match is not None:
                    constraint_labels.add(match.group(1))

            for constraint in constraint_labels:
                lhs = load(zf.read(f"constraints/{constraint}/lhs"))
                rhs = np.frombuffer(zf.read(f"constraints/{constraint}/rhs"), np.float64)[0]
                sense = zf.read(f"constraints/{constraint}/sense").decode('ascii')
                discrete = any(zf.read(f"constraints/{constraint}/discrete"))

                label = deserialize_variable(json.loads(constraint))

                try:
                    weight = np.frombuffer(zf.read(f"constraints/{constraint}/weight"), np.float64)[0]
                    penalty = zf.read(f"constraints/{constraint}/penalty").decode('ascii')
                except KeyError:
                    weight = None
                    penalty = None

                cqm.add_constraint(lhs, rhs=rhs, sense=sense, label=label, weight=weight, penalty=penalty, copy=False)
                if discrete:
                    cqm.discrete.add(label)

        if check_header:
            expected = dict(num_variables=len(cqm.variables),
                            num_constraints=len(cqm.constraints),
                            num_biases=cqm.num_biases(),
                            )
            if header_info.version >= (1, 1):
                expected.update(
                    num_quadratic_variables=cqm.num_quadratic_variables(include_objective=False))
            if header_info.version >= (1, 2):
                expected.update(
                    num_quadratic_variables_real=cqm.num_quadratic_variables(Vartype.REAL, include_objective=True),
                    num_linear_biases_real=cqm.num_biases(Vartype.REAL, linear_only=True),
                    )
            if header_info.version >= (1, 3):
                expected.update(
                    num_weighted_constraints=cqm.num_soft_constraints(),
                    )

            if expected != header_info.data:
                raise ValueError(
                    "header data does not match the deserialized CQM. "
                    f"Expected {expected!r}, recieved {header_info.data!r}"
                    )

        return cqm

    def iter_constraint_data(self, sample_like: SamplesLike) -> Iterator[ConstraintData]:
        """Yield information about the constraints for the given sample.

        Note that this method iterates over constraints in the same order as
        they appear in :attr:`.constraints`.

        Args:
            sample_like: A sample. `sample-like` is an extension of
                NumPy's array_like structure. See :func:`.as_samples`.

        Yields:
            A :class:`collections.namedtuple` with the following fields.

            * ``label``: Constraint label.
            * ``lhs_energy``:  Energy of the left-hand side of the constraint.
            * ``rhs_energy``: Energy of the right-hand side of the constraint.
            * ``sense``: :class:`dimod.sym.Sense` of the constraint.
            * ``activity``: Equals ``lhs_energy - rhs_energy``.
            * ``violation``: Ammount by which the constraint is violated, if
              positive, or satisfied, if negative. Determined by the type of
              constraint.

        Examples:
            The sample in this example sets a value of ``2`` for two constraints,
            :math:`i \le 3` and :math:`j \ge 3`, which satisfies the first and
            violates the second by the same ammount, flipping the sign of the
            ``violation`` field.

            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> i, j = dimod.Integers(["i", "j"])
            >>> cqm.add_constraint_from_comparison(i <= 3, label="Upper limit")
            'Upper limit'
            >>> cqm.add_constraint_from_comparison(j >= 3, label="Lower limit")
            'Lower limit'
            >>> for constraint in cqm.iter_constraint_data({"i": 2, "j": 2}):
            ...     print(constraint.label, constraint.violation)
            Upper limit -1.0
            Lower limit 1.0
        """
        # We go ahead and coerce to Variables for performance, since .energies() prefers
        # that format
        sample, labels = as_samples(sample_like, labels_type=Variables)

        if sample.shape[0] != 1:
            raise ValueError("sample_like should be a single sample, "
                             f"received {sample.shape[0]} samples")

        for label, constraint in self.constraints.items():
            lhs = constraint.lhs.energy((sample, labels))
            rhs = constraint.rhs
            sense = constraint.sense

            activity = lhs - rhs

            if sense is Sense.Eq:
                violation = abs(activity)
            elif sense is Sense.Ge:
                violation = -activity
            elif sense is Sense.Le:
                violation = activity
            else:
                raise RuntimeError("unexpected sense")

            yield ConstraintData(
                activity=activity,
                sense=sense,
                violation=violation,
                lhs_energy=lhs,
                rhs_energy=rhs,
                label=label,
                )

    def iter_violations(self, sample_like: SamplesLike, *,
                        skip_satisfied: bool = False,
                        clip: bool = False,
                        ) -> Iterator[Tuple[Hashable, Bias]]:
        """Yield violations for all constraints.

        Args:
            sample_like: A sample. `sample-like` is an extension of
                NumPy's array_like structure. See :func:`.as_samples`..
            skip_satisfied: If True, does not yield constraints that are satisfied.
            clip: If True, negative violations are rounded up to 0.

        Yields:
            A 2-tuple containing the constraint label and the amount of the
            constraint's violation.

        Example:

            Construct a constrained quadratic model.

            >>> i, j, k = dimod.Binaries(['i', 'j', 'k'])
            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> cqm.add_constraint(i + j + k == 10, label='equal')
            'equal'
            >>> cqm.add_constraint(i + j <= 15, label='less equal')
            'less equal'
            >>> cqm.add_constraint(j - k >= 0, label='greater equal')
            'greater equal'

            Check the violations of a sample that satisfies all constraints.

            >>> sample = {'i': 3, 'j': 5, 'k': 2}
            >>> for label, violation in cqm.iter_violations(sample, clip=True):
            ...     print(label, violation)
            equal 0.0
            less equal 0.0
            greater equal 0.0

            Check the violations for a sample that does not satisfy all of the
            constraints.

            >>> sample = {'i': 3, 'j': 2, 'k': 5}
            >>> for label, violation in cqm.iter_violations(sample, clip=True):
            ...     print(label, violation)
            equal 0.0
            less equal 0.0
            greater equal 3.0

            >>> sample = {'i': 3, 'j': 2, 'k': 5}
            >>> for label, violation in cqm.iter_violations(sample, skip_satisfied=True):
            ...     print(label, violation)
            greater equal 3.0

        """
        if skip_satisfied:
            # clip doesn't matter in this case
            # todo: feasibility tolerance?
            for datum in self.iter_constraint_data(sample_like):
                if datum.violation > 0:
                    yield datum.label, datum.violation
        elif clip:
            for datum in self.iter_constraint_data(sample_like):
                yield datum.label, max(datum.violation, 0.0)
        else:
            for datum in self.iter_constraint_data(sample_like):
                yield datum.label, datum.violation

    def is_almost_equal(self, other: 'ConstrainedQuadraticModel',
                        places: int = 7) -> bool:
        """Test for near equality to a given constrained quadratic model.

        All biases of the objective and constraints are compared.

        Args:
            other:
                Constrained quadratic model with which to compare biases.
            places:
                Number of decimal places to which the Python :func:`round`
                function calculates approximate equality.

        Examples:
            >>> cqm1 = dimod.ConstrainedQuadraticModel.from_quadratic_model(
            ...   dimod.BinaryQuadraticModel({0: 0.1234}, {(0, 1): -1.1234}, 0, "BINARY"))
            >>> cqm2 = dimod.ConstrainedQuadraticModel.from_quadratic_model(
            ...   dimod.BinaryQuadraticModel({0: 0.1232}, {(0, 1): -1.1229}, 0, "BINARY"))
            >>> cqm1.is_almost_equal(cqm2, 4)
            False
            >>> cqm1.is_almost_equal(cqm2, 3)
            True
        """
        def constraint_eq(c0: Comparison, c1: Comparison) -> bool:
            return (c0.sense is c1.sense
                    and c0.lhs.is_almost_equal(c1.lhs, places=places)
                    and not round(c0.rhs - c1.rhs, places))

        return (self.objective.is_almost_equal(other.objective, places=places)
                and self.constraints.keys() == other.constraints.keys()
                and all(constraint_eq(constraint, other.constraints[label])
                        for label, constraint in self.constraints.items()))

    def is_equal(self, other: 'ConstrainedQuadraticModel') -> bool:
        """Test for equality to a given constrained quadratic model.

        All biases of the objective and constraints are compared.

        Args:
            other:
                Constrained quadratic model with which to compare biases.
        """
        def constraint_eq(c0: Comparison, c1: Comparison) -> bool:
            return (c0.sense is c1.sense
                    and c0.lhs.is_equal(c1.lhs)
                    and c0.rhs == c1.rhs)

        return (self.objective.is_equal(other.objective)
                and self.constraints.keys() == other.constraints.keys()
                and all(constraint_eq(constraint, other.constraints[label])
                        for label, constraint in self.constraints.items()))

    def is_linear(self) -> bool:
        """Return True if the model has no quadratic interactions."""
        return (self.objective.is_linear() and
                all(comp.lhs.is_linear() for comp in self.constraints.values()))

    def _new_constraint_label(self) -> str:
        # we support up to 100k constraints and :6 gives us 16777216
        # possible so pretty safe
        label = 'c' + uuid.uuid4().hex[:6]
        while label in self.constraints:
            label = 'c' + uuid.uuid4().hex[:6]
        return label

    def num_biases(self, vartype: Optional[VartypeLike] = None, *,
                   linear_only: bool = False,
                   ) -> int:
        """Number of biases in the constrained quadratic model.

        Includes biases in both the objective and any constraints.

        Args:
            vartype: Count only variables of the specified :class:`~dimod.Vartype`.

            linear_only: Count only linear biases.

        Returns:
            The number of biases.

        Examples:
            This example counts the three linear biases (including a linear bias
            implicitly set to zero for variable ``y``) and two quadratic baises.

            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> i = dimod.Integer("i")
            >>> x, y = dimod.Binaries(["x", "y"])
            >>> cqm.set_objective(2*x + 3*i - x*y + i*x)
            >>> cqm.num_biases()
            5

        """
        if vartype is None:
            def count(qm: Union[QuadraticModel, BinaryQuadraticModel]) -> int:
                return qm.num_variables + (0 if linear_only else qm.num_interactions)
        else:
            vartype = as_vartype(vartype, extended=True)

            def count(qm: Union[QuadraticModel, BinaryQuadraticModel]) -> int:
                if isinstance(qm, BinaryQuadraticModel):
                    if qm.vartype is not vartype:
                        return 0
                    return qm.num_variables + (0 if linear_only else qm.num_interactions)
                else:
                    num_biases = sum(qm.vartype(v) is vartype for v in qm.variables)
                    if not linear_only:
                        num_biases += sum(qm.vartype(u) is vartype or qm.vartype(v) is vartype
                                          for u, v, _ in qm.iter_quadratic())
                    return num_biases

        return count(self.objective) + sum(count(const.lhs) for const in self.constraints.values())

    def num_quadratic_variables(self, vartype: Optional[VartypeLike] = None, *,
                                include_objective: Optional[bool] = None,
                                ) -> int:
        """Number of variables with at least one quadratic interaction in the constrained quadratic model.

        Includes interactions in both the objective and any constraints.

        Args:
            vartype: Count only variables of the specified :class:`~dimod.Vartype`.

            include_objective: Count also variables in the objective. Currently defaults to false.

        Return:
            The number of variables.

        Examples:
            This example counts the two variables participating in interaction
            ``3*i*k`` in constraint ``Constraint1`` but not variable ``j``'s
            interaction ``2*i*j`` in the objective.

            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> i, j, k = dimod.Integers(["i", "j", "k"])
            >>> cqm.set_objective(i - 2*i*j)
            >>> cqm.add_constraint_from_comparison(3*i*k - 2*j <= 4, label="Constraint1")
            'Constraint1'
            >>> cqm.num_quadratic_variables()
            2

        .. deprecated:: 0.10.14

            In dimod 0.12.0 ``include_objective`` will default to true.

        """
        if include_objective is None:
            warnings.warn(
                "in dimod 0.12.0 the default value of include_objective will change to true. "
                "To suppress this warning while keeping the existing behavior, set include_objective=False. "
                "To get the new behavior, set include_objective=True",
                DeprecationWarning,
                stacklevel=2,
                )

        if vartype is None:
            def count(qm: Union[QuadraticModel, BinaryQuadraticModel]) -> int:
                return sum(qm.degree(v) > 0 for v in qm.variables)
        else:
            vartype = as_vartype(vartype, extended=True)

            def count(qm: Union[QuadraticModel, BinaryQuadraticModel]) -> int:
                if isinstance(qm, BinaryQuadraticModel):
                    return sum(qm.degree(v) > 0 for v in qm.variables) if qm.vartype is vartype else 0
                else:
                    return sum(qm.vartype(v) is vartype and qm.degree(v) > 0 for v in qm.variables)

        n = sum(count(const.lhs) for const in self.constraints.values())

        if include_objective:
            n += count(self.objective)

        return n

    def relabel_constraints(self, mapping: Mapping[Hashable, Hashable]):
        """Relabel the constraints.

        Note that this method does not maintain the constraint order.

        Args:
            mapping: Mapping from the old constraint labels to the new.

        Examples:
            >>> x, y, z = dimod.Binaries(['x', 'y', 'z'])
            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> cqm.add_constraint(x + y == 1, label='c0')
            'c0'
            >>> cqm.add_constraint(y*z == 0, label='c1')
            'c1'
            >>> cqm.relabel_constraints({'c1': 'c2'})
            >>> list(cqm.constraints)
            ['c0', 'c2']

        """
        discrete_indices = [self.constraint_labels.index(c) for c in self.discrete]
        self.constraint_labels._relabel(mapping)
        self.discrete.clear()
        self.discrete |= (self.constraint_labels[i] for i in discrete_indices)

    def relabel_variables(self,
                          mapping: Mapping[Variable, Variable],
                          inplace: bool = True,
                          ) -> ConstrainedQuadraticModel:
        """Relabel the variables of the objective and constraints.

        Args:
            mapping: Mapping from the old variable labels to the new.
            inplace: Relabels the model's variables in-place if True, or returns
                a copy of the model with relabeled variables if False.

        Returns:
            A constrained quadratic model. Itself by default, or a copy if
            ``inplace`` is set to False.

        """
        if not inplace:
            return copy.deepcopy(self).relabel_variables(mapping, inplace=True)

        self.variables._relabel(mapping)

        return self

    def remove_constraint(self, label: Hashable, *, cascade: bool = False):
        """Remove a constraint from the model.

        Args:
            label: Label of the constraint to remove.
            cascade: If set to True, also removes any variables found only in the
                removed constraint that contribute no energy to the objective.

        Examples:
            This example also removes variable ``k`` from the model when removing
            the constraint while keeping the variables used in the objective.

            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> i, j, k = dimod.Integers(["i", "j", "k"])
            >>> cqm.set_objective(2 * i - 3 * i * j)
            >>> cqm.add_constraint_from_comparison(i * k - 2 * j <= 4, label="C1")
            'C1'
            >>> cqm.remove_constraint("C1", cascade=True)
            >>> cqm.variables
            Variables(['i', 'j'])
        """
        if cascade:
            # Identify which variables are unique to the constraint we're about
            # to remove.
            # We could pull this down to the C++ level, but the performance
            # is pretty bad regardless so let's keep it simple and do it here.
            # There are also some potential Python performance improvements here
            to_remove: Set[Variable] = set()
            to_remove.update(self.constraints[label].lhs.variables)
            to_remove.difference_update(self.objective.variables)
            for lbl, comp in self.constraints.items():
                if not to_remove:
                    break
                if lbl != label:
                    to_remove.difference_update(comp.lhs.variables)

        super().remove_constraint(label)
        self.discrete.discard(label)

        if cascade and to_remove:
            for v in to_remove:
                self.remove_variable(v)

    def remove_variable(self, v: Variable):
        for label in self.discrete:
            if v in self.constraints[label].variables:
                # todo: support this
                raise ValueError("cannot remove a variable used in a discrete constraint")

        super().remove_variable(v)

    def spin_to_binary(self, inplace: bool = False) -> ConstrainedQuadraticModel:
        """Convert any spin-valued variables to binary-valued.

        Args:
            inplace: If set to False, returns a new constrained quadratic model.
                Otherwise, the constrained quadratic model is modified in-place.

        """
        if not inplace:
            return copy.deepcopy(self).spin_to_binary(inplace=True)

        for v in self.variables:
            if self.vartype(v) is Vartype.SPIN:
                self.change_vartype(Vartype.BINARY, v)

        return self

    def _substitute_self_loops_from_model(self, qm: Union[ConstraintView, ObjectiveView],
                                          mapping: MutableMapping[Variable, Variable]):
        for u in qm.variables:
            vartype = qm.vartype(u)

            # integer and binary variables never have self-loops
            if vartype is Vartype.SPIN or vartype is Vartype.BINARY:
                continue

            try:
                bias = qm.get_quadratic(u, u)
            except ValueError:
                # no self-loop
                continue

            lb = qm.lower_bound(u)
            ub = qm.upper_bound(u)

            if u not in mapping:
                # we've never seen this integer before
                new: Variable = new_variable_label()

                # on the off chance there are conflicts. Luckily self.variables
                # is global accross all constraints/objective so we don't need
                # to worry about accidentally picking something we'll regret
                while new in self.constraints or new in self.variables:
                    new = new_variable_label()

                mapping[u] = new

                self.add_variable(vartype, new, lower_bound=lb, upper_bound=ub)

                # we don't add the constraint yet because we don't want
                # to modify self.constraints
            else:
                new = mapping[u]

            self.add_variable(vartype, new, lower_bound=lb, upper_bound=ub)

            qm.add_quadratic(u, new, bias)
            qm.remove_interaction(u, u)

    def substitute_self_loops(self) -> Dict[Variable, Variable]:
        """Replace any self-loops in the objective or constraints.

        Self-loop :math:`i^2` is removed by introducing a new variable
        :math:`j` with interaction :math:`i*j` and adding constraint
        :math:`j == i`.

        Acts on the objective and constraints in-place.

        Returns:
            Mapping from the integer variable labels to their introduced
            counterparts. The constraint enforcing :math:`j == i` uses
            the same label.

        """
        # dev note: we can cythonize this for better performance

        mapping: Dict[Variable, Variable] = dict()

        self._substitute_self_loops_from_model(self.objective, mapping)

        for comparison in self.constraints.values():
            self._substitute_self_loops_from_model(comparison.lhs, mapping)

        # finally add the constraints for the variables
        for v, new in mapping.items():
            self.add_constraint([(v, 1), (new, -1)], rhs=0, sense='==', label=new)

        return mapping

    def to_file(self, *, spool_size: int = int(1e9)) -> tempfile.SpooledTemporaryFile:
        """Serialize to a file-like object.

        Args:
            spool_size: Defines the `max_size` passed to the constructor of
                :class:`tempfile.SpooledTemporaryFile`. Determines whether
                the returned file-like's contents will be kept on disk or in
                memory.

        Format Specification (Version 1.3):

            This format is inspired by the `NPY format`_

            The first 8 bytes are a magic string: exactly "DIMODCQM".

            The next 1 byte is an unsigned byte: the major version of the file
            format.

            The next 1 byte is an unsigned byte: the minor version of the file
            format.

            The next 4 bytes form a little-endian unsigned int, the length of
            the header data HEADER_LEN.

            The next HEADER_LEN bytes form the header data. This is a
            json-serialized dictionary. The dictionary is exactly:

            .. code-block:: python

                dict(num_variables=len(cqm.variables),
                     num_constraints=len(cqm.constraints),
                     num_biases=cqm.num_biases(),
                     num_quadratic_variables=cqm.num_quadratic_variables(),
                     num_quadratic_variables_real=cqm.num_quadratic_variables('REAL', include_objective=True),
                     num_linear_biases_real=cqm.num_biases('REAL', linear_only=True),
                     )

            it is terminated by a newline character and padded with spaces to
            make the entire length of the entire header divisible by 64.

            The constraint quadratic model data comes after the header. It is
            encoded as a zip file. The zip file will contain one file
            named `objective`, containing the objective as encoded as a file
            view. It will also contain a directory called `constraints`. The
            `constraints` directory will contain one subdirectory for each
            constraint, each containing `lhs`, `rhs` and `sense` encoding
            the `lhs` as a fileview, the `rhs` as a float and the sense
            as a string. Each directory will also contain a `discrete` file,
            encoding whether the constraint represents a discrete variable.
            Weighted constraints also contain a `weight` and `penalty` file,
            encoding the weight and the penalty type for the constraint.

        Format Specification (Version 1.2):
            This format is the same as Version 1.3, except that there are no
            weighted constraints, and thus the `weight` and `penalty` files
            are not present.

        Format Specification (Version 1.1):

            This format is the same as Version 1.2, except that the data dict
            does not have ``num_quadratic_variables_real`` and
            ``num_linear_biases_real``.

        Format Specification (Version 1.0):

            This format is the same as Version 1.1, except that the data dict
            does not have `num_quadratic_variables`.

        .. _NPY format: https://numpy.org/doc/stable/reference/generated/numpy.lib.format.html

        Examples:
            >>> cqm1 = dimod.ConstrainedQuadraticModel()
            >>> x, y = dimod.Binaries(["x", "y"])
            >>> cqm1.set_objective(2 * x * y - 2 * x)
            >>> cqm_file = cqm1.to_file()
            >>> cqm2 = dimod.ConstrainedQuadraticModel.from_file(cqm_file)
            >>> print(cqm2.objective.to_polystring())
            -2*x + 2*x*y
        """
        file = SpooledTemporaryFile(max_size=spool_size)

        data = dict(num_variables=len(self.variables),
                    num_constraints=len(self.constraints),
                    num_biases=self.num_biases(),
                    num_quadratic_variables=self.num_quadratic_variables(include_objective=False),
                    num_quadratic_variables_real=self.num_quadratic_variables(Vartype.REAL, include_objective=True),
                    num_linear_biases_real=self.num_biases(Vartype.REAL, linear_only=True),
                    num_weighted_constraints=sum(comp.lhs.is_soft() for comp in self.constraints.values()),
                    )

        write_header(file, CQM_MAGIC_PREFIX, data, version=(1, 3))

        # write the values
        with zipfile.ZipFile(file, mode='a') as zf:
            with self.objective.to_file(spool_size=int(1e12)) as f:
                # we can avoid a copy by trying to read from the underlying buffer
                obj = f._file.getbuffer() if isinstance(f._file, io.BytesIO) else f.read()
                zf.writestr('objective', obj)
                # in the case we got the underlying buffer, we need to delete our
                # reference to it, otherwise we get an error when the file is closed
                del obj

            for label, constraint in self.constraints.items():
                # put everything in a constraints/label/ directory
                lstr = json.dumps(serialize_variable(label))

                with constraint.lhs.to_file(spool_size=int(1e12)) as f:
                    # we can avoid a copy by trying to read from the underlying buffer
                    lhs = f._file.getbuffer() if isinstance(f._file, io.BytesIO) else f.read()
                    zf.writestr(f'constraints/{lstr}/lhs', lhs)
                    # in the case we got the underlying buffer, we need to delete our
                    # reference to it, otherwise we get an error when the file is closed
                    del lhs

                rhs = np.float64(constraint.rhs).tobytes()
                zf.writestr(f'constraints/{lstr}/rhs', rhs)

                sense = bytes(constraint.sense.value, 'ascii')
                zf.writestr(f'constraints/{lstr}/sense', sense)

                discrete = bytes((label in self.discrete,))
                zf.writestr(f'constraints/{lstr}/discrete', discrete)

                # soft constraints
                if constraint.lhs.is_soft():
                    weight = np.float64(constraint.lhs.weight()).tobytes()
                    penalty = bytes(constraint.lhs.penalty(), 'ascii')
                    zf.writestr(f'constraints/{lstr}/weight', weight)
                    zf.writestr(f'constraints/{lstr}/penalty', penalty)

        file.seek(0)
        return file

    def violations(self, sample_like: SamplesLike, *,
                   skip_satisfied: bool = False,
                   clip: bool = False) -> Dict[Hashable, Bias]:
        """Return a dict of violations for all constraints.

        The dictionary maps constraint labels to the amount each constraint is
        violated. This method is a shortcut for ``dict(cqm.iter_violations(sample))``.

        Args:
            sample_like: A sample. `sample-like` is an extension of
                NumPy's array_like structure. See :func:`.as_samples`..
            skip_satisfied: If True, does not yield constraints that are satisfied.
            clip: If True, negative violations are rounded up to 0.

        Returns:
            A dict of 2-tuples containing the constraint label and the amount of
            the constraint's violation.

        Examples:

            This example meets one constraint exactly, is well within the
            requirement of a second, and violates a third.

            >>> i, j, k = dimod.Binaries(['i', 'j', 'k'])
            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> cqm.add_constraint(i + j + k == 10, label='equal')
            'equal'
            >>> cqm.add_constraint(i + j <= 15, label='less equal')
            'less equal'
            >>> cqm.add_constraint(j - k >= 0, label='greater equal')
            'greater equal'
            >>> sample = {"i": 3, "j": 2, "k": 5}
            >>> cqm.violations(sample)
            {'equal': 0.0, 'less equal': -10.0, 'greater equal': 3.0}
        """
        return dict(self.iter_violations(sample_like, skip_satisfied=skip_satisfied, clip=clip))

    @classmethod
    def from_lp_file(cls,
                     fp: Union[BinaryIO, ByteString],
                     lower_bound_default: Optional[float] = None,
                     upper_bound_default: Optional[float] = None,
                     ) -> ConstrainedQuadraticModel:
        """Create a constrained quadratic model from an LP file.

        Args:
            fp: A file-like or a binary string.
            lower_bound_default: Deprecated. Does nothing.
            upper_bound_default: Deprecated. Does nothing.

        Returns:
            The constrained quadratic model defined by the LP file.

        .. deprecated:: 0.11.0

            This method will be removed in dimod 0.13.0.
            Use :func:`~dimod.lp.load` or :func:`~dimod.lp.loads` instead.

        .. deprecated:: 0.11.0

            The ``lower_bound_default`` and ``upper_bound_default`` keyword
            arguments are deprecated and do nothing.

        """
        from dimod.lp import load, loads

        warnings.warn(
            "this method is deprecated as of dimod 0.11.0 "
            "and will be removed in 0.13.0. "
            "Use dimod.lp.load() or dimod.lp.load() instead.",
            DeprecationWarning, stacklevel=2)

        if isinstance(fp, (str, bytes)) and not os.path.isfile(fp):
            obj = loads(fp)
        else:
            obj = load(fp)

        return obj

    _STR_MAX_DISPLAY_ITEMS = 10

    def __str__(self) -> str:
        vartype_name = {Vartype.SPIN: 'Spin',
                        Vartype.BINARY: 'Binary',
                        Vartype.INTEGER: 'Integer',
                        Vartype.REAL: 'Real'}

        def var_encoder(v):
            return f'{vartype_name[self.vartype(v)]}({v!r})'

        sio = StringIO()

        def render_limited_number(iterable, render_element):
            tail_limit = self._STR_MAX_DISPLAY_ITEMS // 2
            head_limit = self._STR_MAX_DISPLAY_ITEMS - tail_limit
            limited = False
            tail = []

            for k, x in enumerate(iterable):
                assert x is not None

                if k < head_limit:
                    render_element(x)
                else:
                    tail.append(x)

                    if len(tail) > tail_limit:
                        if not limited:
                            sio.write('  ...\n')
                            limited = True
                        tail.pop(0)

            while tail:
                render_element(tail.pop(0))

        def render_constraint(item):
            label, c = item
            sio.write(f'  {label}: ')
            sio.write(c.to_polystring(encoder=var_encoder))
            sio.write('\n')

        def render_bound(v):
            sio.write(f'  {self.lower_bound(v)} <= {var_encoder(v)} <= {self.upper_bound(v)}\n')

        sio.write('Constrained quadratic model: ')
        sio.write(f'{len(self.variables)} variables, ')
        sio.write(f'{len(self.constraints)} constraints, ')
        sio.write(f'{self.num_biases()} biases\n\n')

        sio.write('Objective\n')
        sio.write('  ')
        sio.write(self.objective.to_polystring(encoder=var_encoder))
        sio.write('\n')

        sio.write('\n')
        sio.write('Constraints\n')
        render_limited_number(self.constraints.items(), render_constraint)

        sio.write('\n')
        sio.write('Bounds\n')
        bound_vars = (v for v in self.variables
                      if self.vartype(v) in (Vartype.INTEGER, Vartype.REAL))
        render_limited_number(bound_vars, render_bound)

        return sio.getvalue()


CQM = ConstrainedQuadraticModel


def _qm_to_bqm(qm: QuadraticModel, integers: MutableMapping[Variable, BinaryQuadraticModel],
               ) -> BinaryQuadraticModel:
    # dev note: probably we'll want to make this function or something similar
    # public facing at some point, but right now the interface is pretty weird
    # and it only returns BINARY bqms

    if any(qm.vartype(v) is Vartype.SPIN for v in qm.variables):
        # bqm is BINARY so we want to handle these
        qm = qm.spin_to_binary(inplace=False)

    bqm = BinaryQuadraticModel(Vartype.BINARY)

    for v in qm.variables:
        if v in integers:
            bqm += qm.get_linear(v) * integers[v]
        else:
            bqm.add_linear(v, qm.get_linear(v))

    for u, v, bias in qm.iter_quadratic():
        if u in integers:
            if v in integers:
                bqm += integers[u] * integers[v] * bias
            else:
                bqm += Binary(v) * integers[u] * bias
        elif v in integers:
            bqm += Binary(u) * integers[v] * bias
        else:
            bqm.add_quadratic(u, v, bias)
    bqm.offset += qm.offset

    return bqm


class CQMToBQMInverter:
    """Invert a sample from a binary quadratic model constructed by :func:`cqm_to_bqm`."""
    __slots__ = ('_binary', '_integers')

    def __init__(self,
                 binary: Mapping[Variable, Vartype],
                 integers: Mapping[Variable, BinaryQuadraticModel]):
        self._binary = binary
        self._integers = integers

    def __call__(self, sample: Mapping[Variable, int]) -> Mapping[Variable, int]:
        new = {}

        for v, vartype in self._binary.items():
            if vartype is Vartype.BINARY:
                new[v] = sample[v]
            elif vartype is Vartype.SPIN:
                new[v] = 2*sample[v] - 1
            else:
                raise RuntimeError("unexpected vartype")

        for v, bqm in self._integers.items():
            new[v] = 0
            for u in bqm.variables:
                new[v] += sample[u] * u[1]

        return new

    @classmethod
    def from_dict(cls, doc: Dict[str, Dict[Variable, Any]]) -> CQMToBQMInverter:
        """Construct an inverter from a serialized representation."""

        integers = {}
        for v, variables in doc['integers'].items():
            v = deserialize_variable(v)

            bqm = BinaryQuadraticModel(Vartype.BINARY)
            bqm.add_linear_from((deserialize_variable(u), u[1]) for u in variables)

            integers[v] = bqm

        return cls(
            dict((deserialize_variable(v), as_vartype(vartype))
                 for v, vartype in doc['binary'].items()),
            integers,
            )

    def to_dict(self) -> Dict[str, Dict[Variable, Any]]:
        """Return a json-serializable encoding of the inverter."""
        # todo: in 3.8 we can used TypedDict for the typing
        return dict(
            binary=dict((serialize_variable(v), vartype.name)
                        for v, vartype in self._binary.items()),
            integers=dict((serialize_variable(v), bqm.variables.to_serializable())
                          for v, bqm in self._integers.items()),
            )


# Developer note: This function is *super* ad hoc. In the future, we may want
# A BQM.from_cqm method or similar, but for now I think it makes sense to
# expose that functionality as a function for easier later deprecation.
def cqm_to_bqm(cqm: ConstrainedQuadraticModel, lagrange_multiplier: Optional[Bias] = None,
               ) -> Tuple[BinaryQuadraticModel, CQMToBQMInverter]:
    """Construct a binary quadratic model from a constrained quadratic model.

    Args:
        cqm: A constrained quadratic model. All constraints must be linear
            and all integer variables must have a lower bound of 0.

        lagrange_multiplier: The penalty strength used when converting
            constraints into penalty models. Defaults to 10x the largest
            bias in the objective.

    Returns:
        A 2-tuple containing a binary quadratic model and a function that converts
        samples over the binary quadratic model back into samples for the
        constrained quadratic model.

    Example:

        Start with a constrained quadratic model

        >>> num_widget_a = dimod.Integer('num_widget_a', upper_bound=7)
        >>> num_widget_b = dimod.Integer('num_widget_b', upper_bound=3)
        >>> cqm = dimod.ConstrainedQuadraticModel()
        >>> cqm.set_objective(-3 * num_widget_a - 4 * num_widget_b)
        >>> cqm.add_constraint(num_widget_a + num_widget_b <= 5, label='total widgets')
        'total widgets'

        Convert it to a binary quadratic model and solve it using
        :class:`dimod.ExactSolver`.

        >>> bqm, invert = dimod.cqm_to_bqm(cqm)
        >>> sampleset = dimod.ExactSolver().sample(bqm)

        Interpret the answer in the original variable classes

        >>> invert(sampleset.first.sample)
        {'num_widget_a': 2, 'num_widget_b': 3}

        Note that the inverter is also serializable.

        >>> import json
        >>> newinvert = dimod.constrained.CQMToBQMInverter.from_dict(
        ...     json.loads(json.dumps(invert.to_dict())))
        >>> newinvert(sampleset.first.sample)
        {'num_widget_a': 2, 'num_widget_b': 3}

    """

    from dimod.generators.integer import binary_encoding  # avoid circular import

    bqm = BinaryQuadraticModel(Vartype.BINARY)
    binary: Dict[Variable, Vartype] = {}
    integers: Dict[Variable, BinaryQuadraticModel] = {}

    # add the variables
    for v in cqm.variables:
        vartype = cqm.vartype(v)

        if vartype is Vartype.SPIN or vartype is Vartype.BINARY:
            binary[v] = vartype
        elif vartype is Vartype.INTEGER:
            if cqm.lower_bound(v) != 0:
                raise ValueError("integer variables must have a lower bound of 0, "
                                 f"variable {v} has a lower bound of {cqm.lower_bound(v)}")
            v_bqm = integers[v] = binary_encoding(v, int(cqm.upper_bound(v)))

            if not v_bqm.variables.isdisjoint(bqm.variables):
                # this should be pretty unusual, so let's not bend over backwards
                # to accommodate it.
                raise ValueError("given CQM has conflicting variables with ones "
                                 "generated by dimod.generators.binary_encoding")

            bqm.add_variables_from((v, 0) for v in v_bqm.variables)
        else:
            raise RuntimeError("unexpected vartype")

    for v in binary:
        bqm.add_variable(v)

    # objective, we know it's always a QM
    bqm += _qm_to_bqm(cqm.objective, integers)

    if lagrange_multiplier is None:
        if cqm.constraints and bqm.num_variables:
            max_bias = max(-bqm.linear.min(), bqm.linear.max())
            if not bqm.is_linear():
                max_bias = max(-bqm.quadratic.min(), bqm.quadratic.max(), max_bias)
            if max_bias:
                lagrange_multiplier = 10 * max_bias
            else:
                lagrange_multiplier = 1
        else:
            lagrange_multiplier = 0  # doesn't matter

    for constraint in cqm.constraints.values():
        lhs = _qm_to_bqm(constraint.lhs, integers)
        rhs = constraint.rhs
        sense = constraint.sense

        if not lhs.is_linear():
            raise ValueError("CQM must not have any quadratic constraints")

        if lhs.vartype is Vartype.SPIN:
            lhs = lhs.change_vartype(Vartype.BINARY, inplace=True)

        # at this point we know we have a BINARY bqm

        if sense is Sense.Eq:
            bqm.add_linear_equality_constraint(
                ((v, lhs.get_linear(v)) for v in lhs.variables),
                lagrange_multiplier,
                lhs.offset - rhs,
                )
        elif sense is Sense.Ge:
            bqm.add_linear_inequality_constraint(
                ((v, lhs.get_linear(v)) for v in lhs.variables),
                lagrange_multiplier,
                new_variable_label(),
                constant=lhs.offset,
                lb=rhs,
                ub=np.iinfo(np.int64).max,
                )
        elif sense is Sense.Le:
            bqm.add_linear_inequality_constraint(
                ((v, lhs.get_linear(v)) for v in lhs.variables),
                lagrange_multiplier,
                new_variable_label(),
                constant=lhs.offset,
                lb=np.iinfo(np.int64).min,
                ub=rhs,
                )
        else:
            raise RuntimeError("unexpected sense")

    return bqm, CQMToBQMInverter(binary, integers)


# register fileview loader
load.register(CQM_MAGIC_PREFIX, ConstrainedQuadraticModel.from_file)
